#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import pyvis.network
import networkx as nx
from treelib import Tree
from .datamodel import *

def visualize_dependency_graph(network: Network) -> pyvis.network.Network:
    G = nx.DiGraph()
    for node in network.graph.nodes:
        if isinstance(node, DataplaneTestNode):
            group = 1
        elif isinstance(node, RouteNode):
            group = 2
        elif isinstance(node, ConfigNode):
            group = 3
        else:
            group = 4
        G.add_node(str(node), group=group)
    for node in network.graph.nodes:
        for pred in node.pred:
            G.add_edge(str(pred), str(node))

    nt = pyvis.network.Network(height=800, width=800, notebook=True, directed=True)
    nt.toggle_hide_edges_on_drag(True)
    nt.from_nx(G)
    return nt

def print_dependency_graph_as_tree(tested_nodes: Iterable[DNode]) -> None:
    trees = []

    visited_nodes = set()
    def dfs(node: DNode, parent: DNode = None, tree: Tree = None):
        repr = str(node)
        if parent == None:
            tree = Tree()
            trees.append(tree)
            tree.create_node(repr, repr)
        else:
            id = repr
            while tree.contains(id):
                id = id + "*"
            tree.create_node(id, id, parent=str(parent))

        if node in visited_nodes:
            return

        visited_nodes.add(node)
        for (pred, _) in node.pred:
            dfs(pred, node, tree)

    for node in tested_nodes:
        dfs(node)

    for tree in trees:
        tree.show()


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import annotations
from typing import Callable, List, Tuple

from .dnode import DNode
from .network import Network

class DependencyTemplate:
    """Template to encode one type of legit edges of dependency graph.
    It specifies the range of nodes it cares about (as listed in left_nodes)
    and given one of such node, the mapper computes where does the dependency 
    edge(s) go to/from, depending on the direction.
    @members
        left_nodes: iterable of DNode -> iterable of DNode
        direction: str
        mapper: DNode -> iterable of DNode
    """
    def __init__(self, name, left_nodes, direction, mapper):
        self.name = name
        self.left_nodes = left_nodes
        self.direction = direction
        self.mapper = mapper

class LazyTemplate:
    """Template to encode one type of legit edges of dependency graph.
    It specifies the range of nodes it cares about (as listed in left_nodes)
    and given one of such node, the mapper computes where does the dependency 
    edge(s) go to/from, depending on the direction.
    @members
        filter: DNode -> a boolean indicating that whether this template is applicable for given node
        worker: DNode -> [parent nodes], [dirty nodes to be enqueued], [new nodes to be saved]
    """
    def __init__(self, name, matcher, worker, state_transition, edge_type):
        self.name: str = name
        self.matcher: Callable[[DNode], bool] = matcher
        self.worker: Callable[[Network, DNode], Tuple[List[DNode], List[DNode], List[DNode]]] = worker
        self.state_transition: Callable[[DNode], None] = state_transition
        self.edge_type: str = edge_type

    def default_state_transition(node: DNode) -> None:
        if "init" in node.status:
            node.status.remove("init")

    def null_state_transition(node: DNode) -> None:
        pass



#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import time
from pybatfish.client.session import Session
from .dnode import *

class BatchManager:
    def __init__(self, session: Session) -> None:
        self.bf = session
        self.node_records = []
        self.host_records = []
        self.policies_records = []
        self.direction_records = []
        self.time_cur = .0

    def add_trp_request(self, node: DNode, host: str, policies: List[str], direction: str):
        self.node_records.append(node)
        self.host_records.append(host)
        self.policies_records.append(policies)
        self.direction_records.append(direction)

    def process_batch(self) -> List[DNode]:
        batch_start = time.time()
        # step1: collect all routes for different (device, policy) pairs
        trp_batch = dict()
        for i, node in enumerate(self.node_records):
            host = self.host_records[i]
            polices = self.policies_records[i]
            direction = self.direction_records[i]

            for policy in polices:
                if (host, policy, direction) in trp_batch:
                    trp_batch[(host, policy, direction)].append(node)
                else:
                    trp_batch[(host, policy, direction)] = [node]
        
        # step 2: evaluate export batch
        # send batch question to batfish and update related nodes
        updated_nodes = []
        for (host, policy, direction), nodes in trp_batch.items():
            input_routes = [node.pred_route for node in nodes]
            trp_batch_result = self.bf.q.testRoutePolicies(nodes=host, policies=policy, inputRoutes=input_routes, direction=direction).answer().frame()
            for i, node in enumerate(nodes):
                trp_result = trp_batch_result.iloc[i]
                if trp_result["Action"] == "PERMIT" and "wait_for_trp" in node.status:
                    # updata RA nodes with trace
                    node.trace = trp_result["Trace"]
                    node.status.remove("wait_for_trp")
                    node.status.add("trp_finished")
                    updated_nodes.append(node)
                    # for export trp, update downstream BGP nodes with output route
                    if isinstance(node, BgpAnnouncementNode) and direction == "out":
                        succ_node = node.succ_node
                        if "wait_for_pred_ra" in succ_node.status:
                            succ_node.pred_route = trp_result["Output_Route"]
                            succ_node.status.remove("wait_for_pred_ra")
                            succ_node.status.add("pred_ra_set")
                            updated_nodes.append(succ_node)

        # clean up
        self.node_records.clear()
        self.host_records.clear()
        self.policies_records.clear()
        self.direction_records.clear()

        batch_end = time.time()
        self.time_cur += batch_end - batch_start
        return updated_nodes
        


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import annotations
from collections import defaultdict
from typing import DefaultDict, List, Optional, Dict
import pandas as pd
import numpy as np

from pybatfish.datamodel import *
from pybatfish.datamodel.primitives import FileLines

from .sortedrib import RibRule

class RibFrame:
    def __init__(self, rib_type: str, host: str, vrf: str, frame: pd.DataFrame):
        self.rib_type: str = rib_type
        self.host: str = host
        self.vrf: str = vrf
        self.frame: pd.DataFrame = frame

class IndexedRib:
    def __init__(self, rib_type: str, host: str, vrf: str, frame: pd.DataFrame):
        self.rib_type: str = rib_type
        self.host: str = host
        self.vrf: str = vrf
        self.dtype: np.dtype = frame[0:0].to_records().dtype
        self.prefixmap: Dict[str, np.recarray] = {prefix: group.to_records() for prefix, group in frame.groupby("Network")}

    def select_prefix(self, prefix: str) -> np.recarray:
        return self.prefixmap.get(prefix, np.recarray([], self.dtype))

    def select_rule(self, rule: RibRule) -> np.recarray:
        prefix_matched = self.prefixmap.get(rule.network, np.recarray([], self.dtype))
        return prefix_matched[prefix_matched.Next_Hop == rule.nexthop]

class BgpEdge:
    def __init__(self, sender, receiver, sender_vrf, receiver_vrf, sender_as, receiver_as, sender_ip, receiver_ip, sender_export_policy, receiver_import_policy):
        self.sender:str = sender
        self.receiver: str = receiver
        self.sender_vrf: str = sender_vrf
        self.receiver_vrf: str = receiver_vrf
        self.sender_as: int = sender_as
        self.receiver_as: int = receiver_as
        self.send_ip: str = sender_ip
        self.receiver_ip: str = receiver_ip
        self.sender_export_policy: List[str] = sender_export_policy
        self.receiver_import_policy: List[str] = receiver_import_policy
        # data structure for external bgp announcements
        self.bgp_routes: DefaultDict[str, List[BgpRoute]] = defaultdict(list)

class BgpPeerConfig:
    def __init__(self, host, vrf, is_passive, local_as, export_policies, import_policies, peer_group):
        self.host: str = host
        self.vrf: str = vrf
        self.is_passive: bool = is_passive
        self.local_as: int = local_as
        self.export_policies: List[str] = export_policies
        self.import_policies: List[str] = import_policies
        self.peer_group: str = peer_group
        self.lines: FileLines = None

class BgpPeerConfigPassive(BgpPeerConfig):
    def __init__(self, host, vrf, is_passive, local_as, export_policies, import_policies, peer_group, listen_range):
        super().__init__(host, vrf, is_passive, local_as, export_policies, import_policies, peer_group)
        self.listen_range: str = listen_range

    def __repr__(self) -> str:
        return f"BgpPeerPassiveConfig@{self.host}.{self.vrf} listen range:{self.listen_range}"

class BgpPeerConfigP2p(BgpPeerConfig):
    def __init__(self, host, vrf, is_passive, local_as, export_policies, import_policies, peer_group, local_ip, remote_ip, remote_as):
        super().__init__(host, vrf, is_passive, local_as, export_policies, import_policies, peer_group)
        self.local_ip: str = local_ip
        self.remote_ip: str = remote_ip
        self.remote_as: int = remote_as
    
    def __repr__(self) -> str:
        return f"BgpPeerP2pConfig@{self.host}.{self.vrf} peer:{self.remote_as}@{self.remote_ip}"

class BgpGroupConfigRaw:
    def __init__(self, host, name, typename, lines) -> None:
        self.host: str = host
        self.name: str = name
        self.typename: str = typename
        self.lines: FileLines = lines

class BgpGroupConfig:
    def __init__(self, host, vrf, name, lines, raw_lines) -> None:
        self.host: str = host
        self.vrf: str = vrf
        self.name: str = name
        self.lines: FileLines = lines
        self.raw_lines: FileLines = raw_lines
        self.peer_configs: List[BgpPeerConfig] = []

class Routemap:
    def __init__(self, host: str, name: str, typename: str, lines: FileLines) -> None:
        self.host: str = host
        self.name: str = name
        self.typename: str = typename
        self.lines: FileLines = lines
        self.raw_lines: FileLines = lines
        self.clauses: List[RoutemapClause] = []

    def add_clause(self, clause: RoutemapClause):
        self.clauses.append(clause)

    def get_clause(self, name: str) -> RoutemapClause:
        for clause in self.clauses:
            if clause.name == name:
                return clause

        # special case: default term, which is not captured by definedStructures
        words = name.split("__")
        if "DEFAULT_TERM" in words:
            # cl_lines = {line for line in self.lines.lines}
            # for other_cl in self.clauses:
            #     for line in other_cl.lines.lines:
            #         cl_lines.remove(line)
            default_lines = FileLines(filename=self.lines.filename, lines=self.lines.lines)
            clause = RoutemapClause(self.host, name, "DEFAULT_TERM", default_lines)
            self.clauses.append(clause)
            return clause

        return None

    def __repr__(self) -> str:
        return f"Routemap@{self.host} {self.name}"

class RoutemapClause:
    def __init__(self, host: str, name: str, seq: str, lines: FileLines) -> None:
        self.host: str = host
        self.name: str = name
        self.seq: str = seq    
        self.lines: FileLines = lines 

    def __repr__(self) -> str:
        return f"RM-Clause@{self.host} {self.name}" 


class BgpSessionStatus:
    def __init__(self, host, vrf, peer, local_as, remote_as, local_ip, remote_ip, local_interface, remote_interface, session_type, established_status):
        self.host: str = host
        self.vrf: str = vrf
        self.peer: str = peer
        self.local_as: int = local_as
        self.remote_as: int = remote_as
        self.local_ip: Optional[str] = local_ip
        self.remote_ip: Optional[str] = remote_ip
        self.local_interface: Optional[str] = local_interface
        self.remote_interface: Optional[str] = remote_interface
        self.session_type: str = session_type
        self.established_status: str = established_status
        self.is_border: bool = False

    def __repr__(self) -> str:
        return f"BgpSession@{self.host}.{self.vrf}({self.local_as}, {self.local_ip})-{self.peer}({self.remote_as}, {self.remote_ip})"
    
class InterfaceConfig:
    def __init__(self, host, name, lines) -> None:
        self.host: str = host
        self.vrf: str = "unknown-vrf"
        self.name: str = name
        self.lines: FileLines = lines

    def __repr__(self) -> str:
        return f"Interface[{self.name}]@{self.host}.{self.vrf}" 

class ReferencedConfig:
    def __init__(self, host, config_type, name, lines) -> None:
        self.host: str = host
        self.config_type: str = config_type
        self.name: str = name
        self.lines: FileLines = lines
    
    def __repr__(self) -> str:
        return f"{self.config_type.title()}@{self.host} {self.name}" 

    
    
""" class NextHop:
    def __init__(self, type) -> None:
        self.type = type

class NextHopIp(NextHop):
    def __init__(self, type, ip: str) -> None:
        super().__init__(type)
        self.ip = ip

class NextHopInterface(NextHop):
    def __init__(self, type, hostname: str, interface: str) -> None:
        super().__init__(type)
        self.hostname = hostname
        self.interface = interface """


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
TYPE_NAMES_INTERFACE = ["interface"]
TYPE_NAMES_ROUTEMAP = ["route-map", "policy-statement"]
TYPE_NAMES_ROUTEMAP_CLAUSE = ["route-map entry", "route-map-clause", "policy-statement term"]
TYPE_NAMES_ASPATH = ["as-path", "bgp as-path access-list"]
TYPE_NAMES_COMMUNITY = ["community", "community-list", "standard community-list", "bgp community-list standard"]
TYPE_NAMES_BGP_PEER = ["bgp neighbor"]
TYPE_NAMES_BGP_GROUP = ["bgp group"]
TYPE_NAMES_PREFIXLIST = ["prefix-list", "ipv4 prefix-list", "ip prefix-list"]
TYPE_NAMES_VRF = ["routing-instance", "virtual-router"]

SUPPORTED_CONFIG_TYPES = TYPE_NAMES_INTERFACE\
    + TYPE_NAMES_ROUTEMAP\
    + TYPE_NAMES_ROUTEMAP_CLAUSE\
    + TYPE_NAMES_ASPATH\
    + TYPE_NAMES_COMMUNITY\
    + TYPE_NAMES_PREFIXLIST\
    + TYPE_NAMES_BGP_PEER\
    + TYPE_NAMES_BGP_GROUP
    


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import pandas as pd
import json
import pickle
import logging
import parse
from typing import Dict, List, Iterable

from .utils import extract_bgp_neighbor_ip_vrf
from ..datamodel.dnode import DNode
from ..datamodel.network import *
from ..datamodel.template import LazyTemplate
from ..datamodel.configtypes import *
from .templates import all_lazy_templates


def build_control_plane_datamodel(network: Network):
    """build a custom python datamodel for control plane coverage. Information is
    retrieved via Batfish questions and is organized in a "Network" object.
    """
    if network.inited_cp:
        return

    bf = network.bf
    logger = logging.getLogger(__name__)
    bgpConfig: pd.DataFrame = bf.q.bgpPeerConfiguration(nodes="/^(?!isp_).*/").answer().frame()
    bgpProcess: pd.DataFrame = bf.q.bgpProcessConfiguration().answer().frame()
    bgpSession: pd.DataFrame = bf.q.bgpSessionStatus().answer().frame()

    # devices and vrfs
    nodes: pd.DataFrame = bf.q.nodeProperties().answer().frame()
    for rec in nodes[["Node", "VRFs", "Routing_Policies"]].itertuples():
        device_name = rec.Node
        device = Device(device_name)
        for vrf_name in rec.VRFs:
            vrf = Vrf(device_name, vrf_name)
            device.add_vrf(vrf_name, vrf)
        for policy in rec.Routing_Policies:
            device.raw_policies.append(policy)
        network.devices[device_name] = device

    # devices' filenames
    fileStatus: pd.DataFrame = bf.q.fileParseStatus().answer().frame()
    for rec in fileStatus.itertuples():
        device_name = rec.Nodes[0]
        file_name = rec.File_Name
        network.devicenames[file_name] = device_name
        network.filenames[device_name] = file_name
    
    # vrfs' bgp enable flag
    for rec in bgpProcess[["Node", "VRF"]].itertuples():
        vrf = network.get_vrf(rec.Node, rec.VRF)
        if vrf is not None:
            vrf.bgp_enabled = True

    # vrfs' bgp sessions and configs
    network.cnt_bgp_peer_configs = len(bgpConfig.index)
    for device_name, vrf_name, device, vrf in network.iter_vrfs():
        if vrf.bgp_enabled == False or device.is_virtual:
            continue
        
        # vrf's bgp session status for each peer
        for rec in bgpSession[(bgpSession["Node"] == device_name) & (bgpSession["VRF"] == vrf_name)].itertuples():
            session = BgpSessionStatus(rec.Node, rec.VRF, rec.Remote_Node, rec.Local_AS, rec.Remote_AS, rec.Local_IP,\
                rec.Remote_IP, rec.Local_Interface, rec.Remote_Interface, rec.Session_Type, rec.Established_Status)
            if is_virtual_node(rec.Remote_Node):
                session.is_border = True
            vrf.bgp_sessions.append(session)

        # vrf's bgp peer configurations
        configs = []
        for rec in bgpConfig[(bgpConfig["Node"] == device_name) & (bgpConfig["VRF"] == vrf_name)].itertuples():
            if rec.Is_Passive:
                configs.append(BgpPeerConfigPassive(rec.Node, rec.VRF, rec.Is_Passive, rec.Local_AS, rec.Export_Policy,\
                    rec.Import_Policy, rec.Peer_Group, rec.Remote_IP))
            else:
                configs.append(BgpPeerConfigP2p(rec.Node, rec.VRF, rec.Is_Passive, rec.Local_AS, rec.Export_Policy,\
                    rec.Import_Policy, rec.Peer_Group, rec.Local_IP, rec.Remote_IP, rec.Remote_AS))
        vrf.bgp_peer_configs.extend(configs)

    # bgp edges
    for device_name, vrf_name, device, vrf in network.iter_vrfs():
        if device.is_virtual:
            continue
        for session in vrf.bgp_sessions:
            if session.established_status == "ESTABLISHED":
                if session.is_border:
                    # bgp edges of this vrf from each external peer
                    receiver_config = vrf.find_bgp_peer_config_for_session(session)
                    edge_rx = BgpEdge(session.peer, device_name, "default", vrf_name, session.remote_as, session.local_as,\
                        session.remote_ip, session.local_ip, [], receiver_config.import_policies)
                    network.add_bgp_edge(edge_rx)
                    edge_tx = BgpEdge(device_name, session.peer, vrf_name, "default", session.local_as, session.remote_as,\
                        session.local_ip, session.remote_ip, [], receiver_config.export_policies)
                    network.add_bgp_edge(edge_tx)
                else:
                    # bgp edges from this vrf to each internal peer
                    sender_config = vrf.find_bgp_peer_config_for_session(session)
                    if sender_config == None:
                        logger.warning(f"Missing bgp peer config at {device_name} {vrf_name} for {session}")
                        continue

                    receiver_name = session.peer
                    receiver_device = network.devices[receiver_name]
                    receiver_session = receiver_device.find_bgp_session_with_as_ip(session.local_as, session.local_ip)
                    # attempt to find valid bgp sessions with null local ip in some corner cases 
                    if session.local_ip == None:
                        receiver_session = receiver_device.find_bgp_session_with_as(session.local_as)
                        if receiver_session and receiver_session.remote_ip:
                            session.local_ip = receiver_session.remote_ip
                            
                    if receiver_session == None:
                        logger.warning(f"Missing bgp session at {receiver_name} for {session}")
                        continue
                    receiver_vrf_name = receiver_session.vrf
                    receiver_vrf = receiver_device.get_vrf(receiver_vrf_name)
                    receiver_config = receiver_vrf.find_bgp_peer_config_for_session(receiver_session)
                    if receiver_config == None:
                        logger.warning(f"Missing bgp peer config at {receiver_name} {receiver_vrf_name} for {session}")

                    edge = BgpEdge(device_name, receiver_name, vrf_name, receiver_vrf_name, session.local_as, session.remote_as,\
                        session.local_ip, session.remote_ip, sender_config.export_policies, receiver_config.import_policies)
                    network.add_bgp_edge(edge)

    
    # source lines of defined structures
    structures_fr: pd.DataFrame = bf.q.definedStructures().answer().frame()
    supported_config_types = TYPE_NAMES_INTERFACE\
        + TYPE_NAMES_ROUTEMAP\
        + TYPE_NAMES_ROUTEMAP_CLAUSE\
        + TYPE_NAMES_ASPATH\
        + TYPE_NAMES_COMMUNITY\
        + TYPE_NAMES_BGP_PEER\
        + TYPE_NAMES_BGP_GROUP
        # + TYPE_NAMES_PREFIXLIST

    for rec in structures_fr.itertuples():
        structure_type = rec.Structure_Type
        network.source.add_source_lines(rec.Source_Lines)
        network.typed_source[structure_type].add_source_lines(rec.Source_Lines)
        if structure_type in supported_config_types:
            network.supported_source.add_source_lines(rec.Source_Lines)
        
        device_name = network.devicenames[rec.Source_Lines.filename]
        if structure_type in TYPE_NAMES_INTERFACE:
            device = network.devices[device_name]
            it = InterfaceConfig(device_name, rec.Structure_Name, rec.Source_Lines)
            device.interface_configs[rec.Structure_Name] = it
        elif structure_type in TYPE_NAMES_ROUTEMAP:
            device = network.devices[device_name]
            rm_name = rec.Structure_Name
            rm = Routemap(device_name, rm_name, structure_type, rec.Source_Lines)
            device.routemaps[rm_name] = rm
        elif structure_type in TYPE_NAMES_ROUTEMAP_CLAUSE:
            device = network.devices[device_name]
            cl_name = rec.Structure_Name
            words = cl_name.split(' ')
            if len(words) == 2:
                rm_name = words[0]
                cl_seq = words[1]
                cl = RoutemapClause(device_name, cl_name, cl_seq, rec.Source_Lines)
                device.raw_routemap_clauses[rm_name].append(cl)
            else:
                logger.warning(f"Unable to parse route-map clause/policy-statement term name: {cl_name}")
        elif structure_type in TYPE_NAMES_ASPATH + TYPE_NAMES_COMMUNITY + TYPE_NAMES_PREFIXLIST:
            device = network.devices[device_name]
            config = ReferencedConfig(device_name, structure_type, rec.Structure_Name, rec.Source_Lines)
            device.referenced_configs[(structure_type, rec.Structure_Name)] = config
        elif structure_type in TYPE_NAMES_BGP_PEER:
            device = network.devices[device_name]
            prefix, vrf_name = extract_bgp_neighbor_ip_vrf(rec.Structure_Name)
            is_ipv4, remote_ip = convert_ipv4_prefix(prefix)
            if is_ipv4:
                config = device.find_bgp_peer_with_ip(remote_ip, vrf_name)
                if config is not None:
                    config.lines = rec.Source_Lines
                else:
                    logger.warning(f"Cannot find bgp peer for {remote_ip} at {device_name}")
        elif structure_type in TYPE_NAMES_BGP_GROUP:
            device = network.devices[device_name]
            config = BgpGroupConfigRaw(device_name, rec.Structure_Name, structure_type, rec.Source_Lines)
            device.raw_bgp_groups[rec.Structure_Name] = config
        elif structure_type in TYPE_NAMES_VRF:
            vrf_name = rec.Structure_Name
            vrf = network.get_vrf(device_name, vrf_name)
            if vrf is not None:
                vrf.source_lines.add_source_lines(rec.Source_Lines)
    
    # refine prefix-list sourcelines
    vrf_and_bgp_lines = SourceLines()
    for rec in structures_fr.itertuples():
        if rec.Structure_Type in TYPE_NAMES_BGP_GROUP + TYPE_NAMES_VRF + TYPE_NAMES_BGP_PEER:
            vrf_and_bgp_lines.add_source_lines(rec.Source_Lines)
    
    for key in TYPE_NAMES_PREFIXLIST:
        # reset typed source
        if key in network.typed_source:
            network.typed_source[key] = SourceLines()
    for device_name, device in network.devices.items():
        for (stype, sname), config in device.referenced_configs.items():
            if stype in TYPE_NAMES_PREFIXLIST:
                list_lines = SourceLines()
                list_lines.add_source_lines(config.lines)
                # some prefix-lists are expanded beyond their declarations by "apply-path" statement
                remove_apply_path = list_lines.diff(vrf_and_bgp_lines)
                # add to typed sources
                network.supported_source.update(remove_apply_path)
                network.typed_source[stype].update(remove_apply_path)
                # update config
                filename = network.devicename_to_filename(device_name)
                config.lines = remove_apply_path.to_filelines(filename)


    # refine default vrf sourcelines
    for device_name, device in network.devices.items():
        filename = network.devicename_to_filename(device_name)
        default_vrf_lines: SourceLines = network.source.select_files([filename])
        for vrf_name, vrf in device.vrfs.items():
            if vrf_name != "default":
                vrf_lines = vrf.source_lines
                default_vrf_lines = default_vrf_lines.diff(vrf_lines)
        default_vrf = device.get_vrf("default")
        default_vrf.source_lines = default_vrf_lines

    # refine bgp groups data model
    for key in TYPE_NAMES_BGP_GROUP:
        if key in network.typed_source:
            network.typed_source[key] = SourceLines()
    for device_name, device in network.devices.items():
        for group_name, bgp_group_raw in device.raw_bgp_groups.items():
            group_lines = SourceLines()
            group_lines.add_source_lines(bgp_group_raw.lines)
            # intersect with each vrf
            for vrf_name, vrf in device.vrfs.items():
                lines_within_vrf = group_lines.intersect(vrf.source_lines)
                if lines_within_vrf.count() > 0:
                    common_lines = lines_within_vrf
                    peers = []
                    # subtract peer configs within this group
                    for peer_config in vrf.bgp_peer_configs:
                        if peer_config.peer_group == group_name:
                            peer_lines = SourceLines()
                            peer_lines.add_source_lines(peer_config.lines)
                            common_lines = common_lines.diff(peer_lines)
                            peers.append(peer_config)
                    filename = network.devicename_to_filename(device_name)
                    lines = common_lines.to_filelines(filename)
                    raw_lines = lines_within_vrf.to_filelines(filename)
                    config = BgpGroupConfig(device_name, vrf_name, group_name, lines, raw_lines)
                    config.peer_configs.extend(peers)
                    vrf.bgp_group_configs[group_name] = config
                    network.typed_source[bgp_group_raw.typename].update(common_lines)
        device.raw_bgp_groups.clear()

    # refine route-map and clause data model
    routemaps: pd.DataFrame = bf.q.definedStructures(types='|'.join(TYPE_NAMES_ROUTEMAP)).answer().frame()
    routemap_clauses: pd.DataFrame = bf.q.definedStructures(types='|'.join(TYPE_NAMES_ROUTEMAP_CLAUSE)).answer().frame()
    network.cnt_routemaps = len(routemaps.index)
    network.cnt_routemap_clauses = len(routemap_clauses.index)

    for key in TYPE_NAMES_ROUTEMAP:
        if key in network.typed_source:
            network.typed_source[key] = SourceLines()
    for device_name, device in network.devices.items():
        for rm_name, clauses in device.raw_routemap_clauses.items():
            rm = device.get_routemap(rm_name)
            if rm is  None:
                continue
            for cl in clauses:
                rm.add_clause(cl)
            
            common_lines = SourceLines()
            common_lines.add_source_lines(rm.lines)
            for cl in clauses:
                cl_lines = SourceLines()
                cl_lines.add_source_lines(cl.lines)
                common_lines = common_lines.diff(cl_lines)
            filename = network.devicename_to_filename(device_name)
            lines = common_lines.to_filelines(filename)
            rm.lines = lines
        device.raw_routemap_clauses = {}
        for rm_name, rm in device.routemaps.items():
            network.typed_source[rm.typename].add_source_lines(rm.lines)
        
    # refine interface data model
    interface_fr: pd.DataFrame = bf.q.interfaceProperties().answer().frame()
    network.cnt_interface = 0
    for rec in interface_fr.itertuples():
        device_name = rec.Interface.hostname
        device = network.devices[device_name]
        if device.is_virtual:
            continue
        network.cnt_interface += 1
        interface_name = rec.Interface.interface
        vrf_name = rec.VRF
        vrf = network.get_vrf(device_name, vrf_name)
        config = device.get_interface_config(interface_name)
        if config is not None:
            vrf.interfaces[interface_name] = config
            config.vrf = vrf_name
        else:
            logger.warning(f"Missing interface config for {interface_name} at {device_name}")

    # referenced structrue
    ref_fr: pd.DataFrame = bf.q.referencedStructures(types='|'.join(TYPE_NAMES_ASPATH + TYPE_NAMES_COMMUNITY + TYPE_NAMES_PREFIXLIST)).answer().frame()
    for rec in ref_fr.itertuples():
        device_name = network.devicenames[rec.Source_Lines.filename]
        device = network.devices[device_name]
        for line_number in rec.Source_Lines.lines:
            # assumption: only one structure can be referenced in a policy line
            device.referenced_lines[line_number] = device.referenced_configs[(rec.Structure_Type, rec.Structure_Name)]

    # static analysis on dead code and unmodeled code
    if network.static_analysis:
        unmodeled, dead = dead_code_static_analysis(network)
        network.dead_source = dead
        network.source = network.source.diff(unmodeled)
        network.supported_source = network.supported_source.diff(unmodeled)
        network.reachable_source = network.supported_source.diff(dead)
        #for stype, type_source in network.typed_source.items():
        #    network.typed_source[stype] = type_source.diff(unmodeled).diff(dead)
    else:
        network.reachable_source = network.supported_source

    # pre-computation only needs to be done once
    network.inited_cp = True

def build_data_plane_datamodel(network: Network, ext_ras: List[Dict]=[]):
    if network.inited_dp:
        return

    bf = network.bf
    logger = logging.getLogger("__name__")
    routes: pd.DataFrame = bf.q.routes().answer().frame()
    bgpRoutes: pd.DataFrame = bf.q.bgpRib(status="/.*/").answer().frame()

    # vrfs' ribs
    network.cnt_rib_entry = len(routes.index)
    for device_name, vrf_name, device, vrf in network.iter_vrfs():
        # isp nodes are auxilliary nodes modeled by batfish
        # we do not need to process their ribs
        if device.is_virtual:
            continue

        # vrf's main rib
        main_fr = routes[(routes["Node"] == device_name) & (routes["VRF"] == vrf_name)]
        main_rib = IndexedRib("main", device_name, vrf_name, main_fr)
        vrf.add_rib("main", main_rib)

        # vrf's connected rib (including static and interface-local)
        connected_fr = main_fr[main_fr["Protocol"].isin(["connected", "local", "static"])]
        connected_rib = IndexedRib("connected", device_name, vrf_name, connected_fr)
        vrf.add_rib("connected", connected_rib)

        # vrf's bgp rib
        if vrf.bgp_enabled:
            bgp_fr = bgpRoutes[(bgpRoutes["Node"] == device_name) & (bgpRoutes["VRF"] == vrf_name)]
            bgp_rib = IndexedRib("bgp", device_name, vrf_name, bgp_fr)
            vrf.add_rib("bgp", bgp_rib)

        # parsing main rib into custom sorted structure for lpm
        rib = vrf.rib
        for rec in main_fr.itertuples():
            rib.add_rule(rec.Network, rec.Next_Hop)

    # parse external bgp annoucements for fast lookup
    for ra in ext_ras:
        device_name = ra["dstNode"]
        peer_ip = ra["srcIp"]
        peer_as = ra["asPath"][0][0]
        session = network.devices[device_name].find_bgp_session_with_as_ip(peer_as, peer_ip)
        vrf_name = session.vrf
        border_edge = network.get_bgp_edge(f"isp_{peer_as}", "default", device_name, vrf_name)
        if border_edge is None:
            logger.warning(f"WARNING: cannot find bgp edge for external route annoucement {ra}")
            continue
        prefix = ra["network"]
        route = convert_external_ra(ra)
        border_edge.bgp_routes[prefix].append(route)

    network.inited_dp = True

def load_external_bgp_announcements(filename):
    with open(filename) as infile:
        content = json.load(infile)
        records = content["Announcements"]
        return records
    
def pickle_network(network: Network, filename) -> None:
    with open(filename, 'wb') as outfile:
        pickle.dump(network, outfile)

def unpickle_network(filename: str) -> Network:
    with open(filename, 'rb') as infile:
        return pickle.load(infile)

def ifg_lazy_construction(network: Network, given_nodes: Iterable[DNode]):
    """evaluate dependency templates on given starter nodes toward fixpoint
    """
    stack = [node for node in given_nodes]

    def run_to_fixpoint():
        while stack:
            node = stack.pop()
            dup = network.graph.get_node(node)
            if dup != None:
                node = dup
            if len(node.status) == 0:
                continue
                
            #network.graph.add_node(node)
            matched_templates: List[LazyTemplate] = []
            for tpl in all_lazy_templates:
                if tpl.matcher(node):
                    matched_templates.append(tpl)
            
            for tpl in matched_templates:
                parents, dirty_nodes, new_nodes = tpl.worker(network, node)
                # new nodes: add to graph
                for new_node in new_nodes:
                    # python default behavior for dup inserting: dismiss
                    network.graph.add_node(new_node)
                # parent nodes: add arrows
                is_weak = tpl.edge_type == 'disj'    
                for parent in parents:
                    parent_dedup = network.graph.get_node(parent)
                    if parent_dedup != None:
                        parent = parent_dedup
                    node.add_dependency(parent, is_weak)
                # dirty nodes: enqueue
                #for dirty in dirty_nodes:
                stack.extend(dirty_nodes)
                tpl.state_transition(node)

    for node in given_nodes:
        network.graph.add_node(node)

    while True:
        run_to_fixpoint()
        dirty_nodes = network.bm.process_batch()
        if dirty_nodes:
            stack.extend(dirty_nodes)
        else:
            break

def dead_code_static_analysis(network: Network) -> Tuple[SourceLines, SourceLines]:
    bf = network.bf
    dead = SourceLines()
    # IPv6 related code (subset of dead), we mark them as unsupported
    unmodeled = SourceLines()

    bgp_group_and_vrf_lines = SourceLines()
    bgp_group_and_vrf_regex = "|".join(TYPE_NAMES_BGP_GROUP + TYPE_NAMES_VRF)
    bgp_group_and_vrf_fr: pd.DataFrame = bf.q.definedStructures(types=bgp_group_and_vrf_regex).answer().frame()
    for rec in bgp_group_and_vrf_fr.itertuples():
        bgp_group_and_vrf_lines.add_source_lines(rec.Source_Lines)

    # IPv6 bgp groups
    bgp_group_regex = "|".join(TYPE_NAMES_BGP_GROUP)
    defined_fr: pd.DataFrame = bf.q.definedStructures(types=bgp_group_regex).answer().frame()
    for rec in defined_fr.itertuples():
        group_name = rec.Structure_Name
        # heuristics
        if group_name[-1] == '6':
            unmodeled.add_source_lines(rec.Source_Lines)
            dead.add_source_lines(rec.Source_Lines)
    # IPv6 bgp neighbors
    bgp_peer_regex = "|".join(TYPE_NAMES_BGP_PEER)
    defined_fr: pd.DataFrame = bf.q.definedStructures(types=bgp_peer_regex).answer().frame()
    for rec in defined_fr.itertuples():
        peer_name = rec.Structure_Name
        peer_ip, _ = parse.parse("{} ({})", peer_name)
        if is_ipv6_prefix(peer_ip):
            unmodeled.add_source_lines(rec.Source_Lines)
            dead.add_source_lines(rec.Source_Lines)

    # structures that are defined but unused
    unused_fr: pd.DataFrame = bf.q.unusedStructures().answer().frame()
    for rec in unused_fr.itertuples():
        stype = rec.Structure_Type
        if stype in ["as-path", "community", "standard community-list"]:
            dead.add_source_lines(rec.Source_Lines)
        elif stype == "prefix-list":
            list_lines = SourceLines()
            list_lines.add_source_lines(rec.Source_Lines)
            # some prefix-lists are expanded beyond their declarations by "apply-path" statement
            # we want to remove these expanded lines
            remove_apply_path = list_lines.diff(bgp_group_and_vrf_lines)
            dead.update(remove_apply_path)
            # naming heuristics to discriminate v4 and v6 elements
            if 6 in extract_digits(rec.Structure_Name):
                unmodeled.update(remove_apply_path)
            
        elif stype in ["policy-statement", "route-map", "bgp group"]:
            dead.add_source_lines(rec.Source_Lines)
            # naming heuristics to discriminate v4 and v6 elements
            if 6 in extract_digits(rec.Structure_Name):
                unmodeled.add_source_lines(rec.Source_Lines)

    # policies referenced by only IPv6 bgp neighbors or other ribs (isis, vrf, fib, ...)
    ref_fr: pd.DataFrame = bf.q.referencedStructures(types="policy-statement|route-map").answer().frame()
    for rec in ref_fr.itertuples():
        context = rec.Context
        ref_lines = rec.Source_Lines
        device_name = network.devicenames[rec.Source_Lines.filename]
        device = network.devices[device_name]
        config = device.get_routemap(rec.Structure_Name)
        if context in ["bgp import policy-statement", "bgp export policy-statement"]:
            # if all references are from IPv6 peers, the policy is IPv6 only
            if unmodeled.contains(ref_lines):
                unmodeled.add_source_lines(config.lines)
                for cl in config.clauses:
                    unmodeled.add_source_lines(cl.lines)
            # if all references are from dead peers, the policy is dead
            if dead.contains(ref_lines):
                dead.add_source_lines(config.lines)
                for cl in config.clauses:
                    dead.add_source_lines(cl.lines)
        else:
            # other policy usages: isis export, vrf import, fib export, ...
            # mark as unsupported
            unmodeled.add_source_lines(config.lines)
            for cl in config.clauses:
                    unmodeled.add_source_lines(cl.lines)
            dead.add_source_lines(config.lines)
            for cl in config.clauses:
                    dead.add_source_lines(cl.lines)

    # prefix-list, as-path, communities referenced by only IPv6 policies
    ref_fr: pd.DataFrame = bf.q.referencedStructures(types="as-path|community|prefix-list|community-list").answer().frame()
    for rec in ref_fr.itertuples():
        context = rec.Context
        ref_lines = rec.Source_Lines

        # if all references are from IPv6 policies, the config is IPv6
        if unmodeled.contains(ref_lines):
            device_name = network.devicenames[rec.Source_Lines.filename]
            device = network.devices[device_name]
            config = device.referenced_configs[(rec.Structure_Type, rec.Structure_Name)]
            unmodeled.add_source_lines(config.lines)
        # if all references are from dead policies, the config is dead
        if dead.contains(ref_lines):
            device_name = network.devicenames[rec.Source_Lines.filename]
            device = network.devices[device_name]
            config = device.referenced_configs[(rec.Structure_Type, rec.Structure_Name)]
            dead.add_source_lines(config.lines)

    dead = dead.diff(unmodeled)
    dead = dead.intersect(network.supported_source)
    unmodeled = unmodeled.intersect(network.supported_source)
    return unmodeled, dead

        
        
        


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import absolute_import
import os
from contextlib import contextmanager
import subprocess
from pybatfish.client.session import Session
from pybatfish.datamodel.answer import Answer

from .visual import print_dependency_graph_as_tree
from .datamodel.network import Network
from .algorithm.construct import *
from .algorithm.converttrace import *
from .algorithm.coverage import *

ROUTES_COLUMNS = ['Node', 'VRF', 'Network', 'Next_Hop', 'Next_Hop_IP',
       'Next_Hop_Interface', 'Protocol']
BGP_ROUTES_COLUMNS = ['AS_Path', 'Node', 'VRF', 'Network', 'Status', 'Next_Hop', 'Next_Hop_IP',
       'Next_Hop_Interface', 'Protocol', 'Origin_Protocol', 'Received_From_IP']
USER_CONFIG_KEYS = ['host', 'lines']

class Coverage:
    def __init__(self, session: Session, snapshot_path: str, static_analysis: bool=False, prebuilt_model:Optional[str]=None) -> None:
        self.model: Network = Network(session, snapshot_path, static_analysis)
        self.trace: Set[DNode] = set()
        self.is_active = True
        self.latest_result = None
        if prebuilt_model is None:
            self._init_control_plane_datamodel()
        else:
            self.load_model(prebuilt_model)

    def _init_control_plane_datamodel(self) -> None:
        status_before = self.is_active
        self.is_active = False
        build_control_plane_datamodel(self.model)
        self.is_active = status_before

    def _init_data_plane_datamodel(self) -> None:
        status_before = self.is_active
        self.is_active = False
        external_ras_filename = os.path.join(self.model.snapshot_path, "external_bgp_announcements.json")
        external_ras = load_external_bgp_announcements(external_ras_filename) if os.path.exists(external_ras_filename) else []
        build_data_plane_datamodel(self.model, external_ras)
        self.is_active = status_before

    def result(self, metrics: List[str]=[]) -> None:
        status_before = self.is_active
        self.is_active = False
        tested_nodes = list(self.trace)
        if not self.model.inited_dp and any([is_data_plane_trace(node) for node in tested_nodes]):
            self._init_data_plane_datamodel()
        ifg_lazy_construction(self.model, tested_nodes)

        if "vanilla" in metrics or len(metrics) == 0:
            covered_lines = control_plane_coverage(self.model, tested_nodes)
        
        sub_metrics = [metric for metric in metrics if metric in ["weak", "strong"]] 
        if len(sub_metrics) > 0:
            weak_coverage(self.model, tested_nodes, sub_metrics, enable_stats=True)

        self.is_active = status_before
        self.latest_result = covered_lines

    def html_report(self, lcov_path: Optional[str] = None, html_path: Optional[str] = None) -> None:
        if lcov_path is None:
            lcov_path = os.path.join("coverage", "lcov.info")
        if html_path is None:
            html_path = os.path.join("coverage", "HTML_REPORT")
        
        if self.latest_result == None:
            self.result()
        logging.getLogger('netcov').warning(f"Writing coverage report to {os.path.join(self.model.snapshot_path, html_path)}")
        dump_lcov(self.latest_result, self.model.reachable_source, os.path.join(self.model.snapshot_path, lcov_path), True)
        p = subprocess.Popen(["genhtml", lcov_path, "--output-directory", html_path, "--no-function-coverage", "--title", "NetCov", "--legend"], cwd=self.model.snapshot_path, stdout=subprocess.DEVNULL)
        p.wait()

    def collect_trace(self, answer: Answer) -> None:
        if self.is_active:
            if answer['question']['class'] == 'org.batfish.question.traceroute.TracerouteQuestion':
                self.trace.update(convert_traceroute_traces([answer.frame()]))
            elif answer['question']['class'] == 'org.batfish.question.testroutepolicies.TestRoutePoliciesQuestion':
                self.trace.update(convert_trp_traces(self.model, [answer.frame()]))
            elif answer['question']['class'] == 'org.batfish.question.routes.RoutesQuestion' and answer['question']['rib'] == 'MAIN':
                self.trace.update(convert_main_rib_routes(answer.frame()))
            elif answer['question']['class'] == 'org.batfish.question.routes.RoutesQuestion' and answer['question']['rib'] == 'BGP':
                self.trace.update(convert_bgp_routes(answer.frame()))

    def add_tested_routes(self, frame: pd.DataFrame) -> None:
        if all(col in frame.columns for col in BGP_ROUTES_COLUMNS):
            self.trace.update(convert_bgp_routes(frame))
        elif all(col in frame.columns for col in ROUTES_COLUMNS):
            self.trace.update(convert_main_rib_routes(frame))
        else:
            logging.getLogger(__name__).warning(f"WARNING: unrecognized DataFrame")

    def add_tested_configs(self, elements: List[Dict]) -> None:
        for element in elements:
            if not all(key in element for key in USER_CONFIG_KEYS):
                logging.getLogger(__name__).warning(f"WARNING: incomplete config info {element}")
                continue
            element['filename'] = self.model.devicename_to_filename(element['host'])
            self.trace.update(convert_raw_config(element))

    @contextmanager
    def track(self, reset: bool = True):
        try:
            self.resume()
            yield None
        finally:
            self.pause()
            self.result()
            if reset:
                self.reset()

    @contextmanager
    def no_cov(self):
        try:
            temp = self.is_active
            self.is_active = False
            yield None
        finally:
            self.is_active = temp

    # Clear coverage trace
    def reset(self) -> None:
        self.trace.clear()

    # Pause automatic coverage tracking
    def pause(self) -> None:
        self.is_active = False
    
    # Resume automatic coverage tracking
    def resume(self) -> None:
        self.is_active = True

    def treevis(self, nodes: Optional[List[DNode]]=None) -> None:
        input_nodes = list(self.trace) if nodes is None else nodes
        print_dependency_graph_as_tree(input_nodes)

    def save_model(self, filename: str) -> None:
        model = self.model.__dict__
        state_dict = {}
        state_keys = Network.state_keys()
        for key in state_keys:
            if key in model:
                state_dict[key] = model[key]
        with open(filename, 'wb') as outfile:
            pickle.dump(state_dict, outfile)

    def load_model(self, filename: str) -> None:
        with open(filename, 'rb') as infile:
            state_dict = pickle.load(infile)
        self.model.load_state_dict(state_dict)
            


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
from .utils import *
from ..datamodel.network import *
from ..datamodel.dnode import *
from ..datamodel.template import LazyTemplate

def matcher_tested_from_main(node: DNode) -> bool:
    if isinstance(node, DataplaneTestNode) and "init" in node.status:
        return True
    return False

def worker_tested_from_main(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    derived = []

    # infer vrf from interface
    node: DataplaneTestNode = node
    interface_name = node.nexthop_interface
    device = network.devices[node.host]
    config = device.get_interface_config(interface_name)
    if config is not None:
        vrf_name = config.vrf
    else:
        logging.getLogger(__name__).warning(f"WARNING: cannot infer vrf for {node}")
        return [], [], []

    # match vrf's rib
    vrf = network.get_vrf(node.host, vrf_name)
    if vrf is None:
        logging.getLogger(__name__).warning(f"WARNING: cannot find vrf object for {node}")
        return [], [], []
    main_rib = vrf.get_rib("main")
    
    if node.test_type == "forwarded_to_ip":
        prefix = node.prefix
        next_hop_ip = node.nexthop_ip
        prefix_matched = main_rib.select_prefix(prefix)
        nh_matched = prefix_matched[prefix_matched.Next_Hop_IP == next_hop_ip]
        if len(nh_matched) == 1:
            rec = nh_matched[0]
        elif len(nh_matched) == 0:
            if len(prefix_matched) == 1:
                rec = prefix_matched[0]
                logging.getLogger(__name__).warning(f"WARNING: expect next hop ip match for {node}, actual matches: {rec.Next_Hop_IP}")
            else:
                rec = None
        else:
            rec = nh_matched[-1]
            logging.getLogger(__name__).warning(f"WARNING: expect unique match for {node}, actual matches: {[MainRouteNode.from_rec(rec) for rec in nh_matched]}")
        if rec:
            derived.append(MainRouteNode.from_rec(rec))
        else:
            logging.getLogger(__name__).warning(f"WARNING: cannot find rib rule for {node}")
        # find the rib rules that are used to resolve the interface for nexthop ip
        rules = vrf.resolve_rib_rules_for_ip(next_hop_ip)
        for rule in rules:
            rule_matched_records = main_rib.select_rule(rule)
            if len(rule_matched_records) == 1:
                rec = rule_matched_records[0]
                derived.append(MainRouteNode.from_rec(rec))
            else:
                logging.getLogger(__name__).warning(f"WARNING: expect unique match for {rule}, actual matches: {[MainRouteNode.from_rec(rec) for rec in rule_matched_records]}")
        
    elif node.test_type == "forwarded_to_interface":
        interface_name = node.nexthop_interface
        rules = vrf.resolve_rib_rules_for_ip(node.dst_ip)
        for rule in rules:
            rule_matched_records = main_rib.select_rule(rule)
            if len(rule_matched_records) == 1:
                rec = rule_matched_records[0]
                derived.append(MainRouteNode.from_rec(rec))
            else:
                logging.getLogger(__name__).warning(f"WARNING: expect unique match for {rule}, actual matches: {[MainRouteNode.from_rec(rec) for rec in rule_matched_records]}")
    return derived, derived, derived

lazy_tested_from_main = LazyTemplate(
    name="tested_from_main",
    matcher=matcher_tested_from_main,
    worker=worker_tested_from_main,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

def matcher_main_from_bgp(node: DNode) -> bool:
    if isinstance(node, MainRouteNode) and node.protocol in ["bgp", "aggregate", "ibgp"] and "init" in node.status:
        return True
    return False

def worker_main_from_bgp(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    vrf = network.get_vrf(node.host, node.vrf)
    if vrf is None:
        logging.getLogger(__name__).warning(f"WARNING: cannot find vrf object for {node}")
        return [], [], []
    if vrf.bgp_enabled is False:
        logging.getLogger(__name__).warning(f"WARNING: unknown pred for {node} at {vrf.name} where BGP is disabled")
        return [], [], []

    node: MainRouteNode = node
    prefix = node.prefix
    nexthop = node.nexthop
    bgp_rib = vrf.get_rib("bgp")
    prefix_matched = bgp_rib.select_prefix(prefix)
    bgp_matched_records = prefix_matched[(prefix_matched.Next_Hop_IP == nexthop) | (prefix_matched.Next_Hop_Interface == nexthop)]
    derived = []
    for rec in bgp_matched_records:
        if rec.Status[0] == "BEST":
            derived.append(BgpRouteNode.from_rec(rec))

    if len(derived) != 1:
        logging.getLogger(__name__).warning(f"WARNING: expect unique match for {node}, actual matches: {derived}")
    return derived, derived, derived

lazy_main_from_bgp = LazyTemplate(
    name="main_from_bgp",
    matcher=matcher_main_from_bgp,
    worker=worker_main_from_bgp,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

def matcher_main_from_connected(node: DNode) -> bool:
    if isinstance(node, MainRouteNode) and node.protocol in ["connected", "static", "local"] and "init" in node.status:
        return True
    return False

def worker_main_from_connected(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    vrf = network.get_vrf(node.host, node.vrf)
    if vrf is None:
        logging.getLogger(__name__).warning(f"WARNING: cannot find vrf object for {node}")
        return [], [], []

    node: MainRouteNode = node
    prefix = node.prefix
    nexthop = node.nexthop
    connected_rib = vrf.get_rib("connected")
    prefix_matched = connected_rib.select_prefix(prefix)
    connected_matched_records = prefix_matched[(prefix_matched.Next_Hop_IP == nexthop) | (prefix_matched.Next_Hop_Interface == nexthop)]
    derived = []
    if len(connected_matched_records) == 1:
        rec = connected_matched_records[0]
        derived.append(ConnectedRouteNode.from_rec(rec))

    if len(derived) != 1:
        logging.getLogger(__name__).warning(f"WARNING: expect unique match for {node}, actual matches: {derived}")
    return derived, derived, derived

lazy_main_from_connected = LazyTemplate(
    name="main_from_connected",
    matcher=matcher_main_from_connected,
    worker=worker_main_from_connected,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

def matcher_bgp_from_received_ra(node: DNode) -> bool:
    if isinstance(node, BgpRouteNode) and node.origin_protocol in ["bgp", "ibgp"] and "init" in node.status:
        return True
    return False

def worker_bgp_from_received_ra(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    vrf = network.get_vrf(node.host, node.vrf)
    if vrf is None:
        logging.getLogger(__name__).warning(f"WARNING: cannot find vrf object for {node}")
        return [], [], []

    derived = []
    dirty = set()
    node: BgpRouteNode = node
    
    # guess sender-peer according to received_from_ip, which is more reliable than nexthop_ip
    # if non-exist, use nexthop_interface
    nexthop = node.received_from_ip
    matched_sessions: List[BgpSessionStatus] = []
    if nexthop != None:
        for session in vrf.bgp_sessions:
            if session.remote_ip == nexthop or (session.local_interface and session.local_interface.interface == nexthop):
                if session.peer != None:
                    matched_sessions.append(session)
    
    if len(matched_sessions) != 1:
        logging.getLogger(__name__).warning(f"WARNING: ambiguous or missing peer for bgp route {node}: {matched_sessions}")
        return [], [], []

    session = matched_sessions[0]
    
    if session.is_border:
        # find external route annoucement at import ra rib
        # TODO: find pred route with given prefix and as path
        bgp_edge = network.get_bgp_edge(session.peer, "default", session.host, session.vrf)
        routes = bgp_edge.bgp_routes[node.prefix]

        if len(routes) == 0 and node.prefix == "0.0.0.0/0" and is_isp(session.peer):
            route = default_route_from_isp(session)
            bgp_edge.bgp_routes[node.prefix].append(route)
            routes = [route]
        if len(routes) > 1 or len(routes) == 0:
            logging.getLogger(__name__).warning(f"WARNING: expect unique external RA for {node}, actual matches: {len(routes)}")
        
        pred_ra = routes[0]
        
        if bgp_edge.receiver_import_policy:
            # find the composed import policy, if it exists
            # this is a vendor-specific behavior
            device = network.devices[node.host]
            policy = find_composed_peer_policy(device.raw_policies, session.remote_ip, "IMPORT")
            if policy == None:
                # plan B: use original import policy if there is only one
                if len(bgp_edge.receiver_import_policy) == 1:
                    policy = bgp_edge.receiver_import_policy[0]
                else:
                    logging.getLogger(__name__).warning(f"WARNING: ambiguous or missing import policy for bgp route {node}, got: {bgp_edge.receiver_import_policy}")
                    return [], [], []

            node.pred_route = pred_ra
            node.status.add("pred_ra_set")
            node.import_policy = policy
            dirty.add(node) 
        ra_node = BgpAnnouncementNode("route_announcement", node.prefix, nexthop, session.peer, session.host,\
            "default", session.vrf, pred_ra, bgp_edge, None, node)
        
        # status consumed by "bgp_from_border_bgp_session"
        node.status.add("received_from_external")
        node.from_session = session
        dirty.add(node)
    else:
        # find route at peer's bgp
        peer_session = find_peer_session(network, session)
        if peer_session is None:
            logging.getLogger(__name__).warning(f"WARNING: cannot find sender bgp session for bgp route {node} in session {session}")
            return [], [], []

        bgp_edge = network.get_bgp_edge(peer_session.host, peer_session.vrf, session.host, session.vrf)
        peer_vrf = network.get_vrf(peer_session.host, peer_session.vrf)
        #peer_bgp_fr = peer_vrf.get_rib("bgp").frame
        #peer_bgp_records = peer_bgp_fr[peer_bgp_fr["Network"] == node.prefix].to_records()
        #peer_bgp_records = peer_bgp_fr.query("Network == @node.prefix").to_records()
        #peer_bgp_records = peer_bgp_fr.loc[[node.prefix]].to_records()
        peer_bgp_rib = peer_vrf.get_rib("bgp")
        peer_bgp_records = peer_bgp_rib.select_prefix(node.prefix)
        matched_records = []
        for rec in peer_bgp_records:
            if rec.Status and rec.Status[0] == "BEST":
                matched_records.append(rec)

        if len(matched_records) == 0:
            logging.getLogger(__name__).warning(f"WARNING: missing BGP route {node.prefix} at {peer_vrf.name}")
            # corner case: in Batfish some bgp route redistributed from static is not present
            #peer_connected_fr = peer_vrf.get_rib("connected").frame
            #peer_connected_records = peer_connected_fr[peer_connected_fr["Network"] == node.prefix].to_records()
            peer_connected_rib = peer_vrf.get_rib("connected")
            peer_connected_records = peer_connected_rib.select_prefix(node.prefix)
            if len(peer_connected_records) > 0:
                rec = peer_connected_records[0]
                new_route = {"Node": rec.Node, "VRF": "default", "Network": rec.Network, "Status": ["BEST"], "Next_Hop_IP": rec.Next_Hop_IP,\
                    "Next_Hop_Interface": rec.Next_Hop_Interface, "Protocol": "bgp", "AS_Path": "", "Metric": 0,\
                    "Local_Pref": 0, "Communities": [], "Origin_Protocol": "connected", "Origin_Type": "igp",\
                    "Originator_Id": "", "Received_From_IP": None, "Cluster_List": None, "Weight": 32768, "Tag": None}
                new_rec = pd.DataFrame.from_dict([new_route]).to_records()
                peer_bgp_rib.prefixmap[node.prefix] = new_rec
                #peer_bgp_fr = peer_bgp_fr.append(new_df, ignore_index=True)
                #peer_vrf.get_rib("bgp").frame = peer_bgp_fr
                logging.getLogger(__name__).warning(f"WARNING: add BGP route redistrubited from static route for consistency. {node.prefix} at {peer_vrf.name}")
                #peer_bgp_records = peer_bgp_fr[peer_bgp_fr["Network"] == node.prefix].to_records()
                peer_bgp_records = peer_bgp_rib.select_prefix(node.prefix)
                for rec in peer_bgp_records:
                    if rec.Status and rec.Status[0] == "BEST":
                        matched_records.append(rec)
            else:
                return [], [], []

        pred_bgp_route = convert_bgp_route(matched_records[0])
        # The import policies for this bgp route can't be evaluated until its pred RA is complete.
        if bgp_edge.receiver_import_policy:
            # find the composed import policy, if it exists
            # this is a vendor-specific behavior
            device = network.devices[node.host]
            policy = find_composed_peer_policy(device.raw_policies, session.remote_ip, "IMPORT")
            if policy == None:
                # plan B: use original import policy if there is only one
                if len(bgp_edge.receiver_import_policy) == 1:
                    policy = bgp_edge.receiver_import_policy[0]
                else:
                    logging.getLogger(__name__).warning(f"WARNING: ambiguous or missing import policy for bgp route {node}, got: {bgp_edge.receiver_import_policy}")
                    return [], [], []

            # state machine:
            # init -> wait_for_pred_ra -> pred_ra_set -> wait_for_trp -> trp_finished
            node.status.add("wait_for_pred_ra")
            node.import_policy = policy
            dirty.add(node)
        # The full data fields of the RA will be filled later by TRP batching
        # For now, we instantiate an incomplete RA node with known prefix, pred bgp route, etc.
        ra_node = BgpAnnouncementNode("route_announcement", node.prefix, nexthop, peer_session.host, session.host,\
            peer_session.vrf, session.vrf, pred_bgp_route, bgp_edge, None, node)
        # state consumed by "sent_ra_from_bgp"
        ra_node.status.add("sent_by_known_device")

    derived.append(ra_node)
    dirty.add(ra_node)

    if len(derived) != 1:
        logging.getLogger(__name__).warning(f"WARNING: expect unique match for {node}, actual matches: {derived}")
    return derived, list(dirty), derived

lazy_bgp_from_received_ra = LazyTemplate(
    name="bgp_from_received_ra",
    matcher=matcher_bgp_from_received_ra,
    worker=worker_bgp_from_received_ra,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

def matcher_bgp_aggregated_from_subnets(node: DNode) -> bool:
    if isinstance(node, BgpRouteNode) and node.origin_protocol == "aggregate" and "init" in node.status:
        return True
    return False

def worker_bgp_aggregated_from_subnets(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    vrf = network.get_vrf(node.host, node.vrf)
    if vrf is None:
        logging.getLogger(__name__).warning(f"WARNING: cannot find vrf object for {node}")
        return [], [], []

    derived = []
    node: BgpRouteNode = node

    agg_prefix = node.prefix
    agg_network = ipaddress.IPv4Network(agg_prefix)

    bgp_rib = vrf.get_rib("bgp")
    for prefix, records in bgp_rib.prefixmap.items():
        route_network = ipaddress.IPv4Network(prefix)
        if route_network.subnet_of(agg_network):
            if prefix == agg_prefix:
                continue
            for rec in records:
                if rec.Status[0] == "BEST":
                    derived.append(BgpRouteNode.from_rec(rec))
        
    return derived, derived, derived

lazy_bgp_aggregated_from_subnets = LazyTemplate(
    name="bgp_aggregated_from_subnets",
    matcher=matcher_bgp_aggregated_from_subnets,
    worker=worker_bgp_aggregated_from_subnets,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="disj"
)

def matcher_request_export_trp_for_ra(node: DNode) -> bool:
    if isinstance(node, BgpAnnouncementNode) and "init" in node.status:
        return True
    return False

def worker_request_export_trp_for_ra(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    # decide whether a trp is needed to compute trace of export policy
    node: BgpAnnouncementNode = node
    bgp_edge = node.bgp_edge
    export_policies = bgp_edge.sender_export_policy

    if export_policies:
        # send route to trp batch
        # use policy combination instead of individual policies

        # find the composed import policy, if it exists
        # this is a vendor-specific behavior
        device = network.devices[node.sender]
        policy = find_composed_peer_policy(device.raw_policies, bgp_edge.receiver_ip, "EXPORT")
        if policy == None:
            # plan B: use original import policy if there is only one
            if len(export_policies) == 1:
                policy = export_policies[0]
            else:
                logging.getLogger(__name__).warning(f"WARNING: ambiguous or missing export policy for {node}, got: {export_policies}")
                return [], [], []
        
        network.bm.add_trp_request(node, bgp_edge.sender, [policy], "out")

        # state machine of RA nodes:
        # init -> wait_for_trp -> trp_finished -> done
        node.status.add("wait_for_trp")

    return [], [], []

lazy_request_export_trp_for_ra = LazyTemplate(
    name="request_export_trp_for_ra",
    matcher=matcher_request_export_trp_for_ra,
    worker=worker_request_export_trp_for_ra,
    state_transition=LazyTemplate.null_state_transition,
    edge_type="conj"
)

def matcher_request_import_trp_for_bgp(node: DNode) -> bool:
    if isinstance(node, BgpRouteNode) and "pred_ra_set" in node.status:
        return True
    return False

def worker_request_import_trp_for_bgp(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    node: BgpRouteNode = node
    network.bm.add_trp_request(node, node.host, [node.import_policy], "in")
    node.status.add("wait_for_trp")

    return [], [], []

def state_transition_request_import_trp_for_bgp(node: DNode):
    if "pred_ra_set" in node.status:
        node.status.remove("pred_ra_set")

lazy_request_import_trp_for_bgp = LazyTemplate(
    name="request_import_trp_for_bgp",
    matcher=matcher_request_import_trp_for_bgp,
    worker=worker_request_import_trp_for_bgp,
    state_transition=state_transition_request_import_trp_for_bgp,
    edge_type="conj"
)

def matcher_bgp_from_connected(node: DNode) -> bool:
    if isinstance(node, BgpRouteNode) and node.origin_protocol == "connected" and "init" in node.status:
        return True
    return False

def worker_bgp_from_connected(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    node: BgpRouteNode = node
    vrf = network.get_vrf(node.host, node.vrf)
    if vrf is None:
        logging.getLogger(__name__).warning(f"WARNING: cannot find vrf object for {node}")
        return [], [], []

    prefix = node.prefix
    nexthop = node.nexthop
    rib = vrf.get_rib("connected")
    prefix_matched = rib.select_prefix(prefix)
    connected_matched_records = prefix_matched[(prefix_matched.Next_Hop_IP == nexthop) | (prefix_matched.Next_Hop_Interface == nexthop)]
    derived = []
    if len(connected_matched_records) == 1:
        rec = connected_matched_records[0]
        derived.append(ConnectedRouteNode.from_rec(rec))
    
    if len(derived) != 1:
        logging.getLogger(__name__).warning(f"WARNING: expect unique match for {node}, actual matches: {derived}")
    return derived, derived, derived

lazy_bgp_from_connected = LazyTemplate(
    name="bgp_from_connected",
    matcher=matcher_bgp_from_connected,
    worker=worker_bgp_from_connected,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

def matcher_sent_ra_from_bgp(node: DNode) -> bool:
    if isinstance(node, BgpAnnouncementNode) and "init" in node.status and "sent_by_known_device" in node.status:
        return True
    return False

def worker_sent_ra_from_bgp(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    node: BgpAnnouncementNode = node
    vrf = network.get_vrf(node.sender, node.sender_vrf)
    if vrf is None:
        logging.getLogger(__name__).warning(f"WARNING: cannot find vrf object for {node}")
        return [], [], []

    prefix = node.prefix
    #bgp_fr = vrf.get_rib("bgp").frame
    #gp_perfix_matched_records = bgp_fr[bgp_fr["Network"] == prefix].to_records()
    #bgp_perfix_matched_records = bgp_fr.query("Network == @prefix").to_records()
    #bgp_perfix_matched_records = bgp_fr.loc[[prefix]].to_records()
    #bgp_perfix_matched_records = bgp_fr.loc[[prefix]]
    bgp_rib = vrf.get_rib("bgp")
    bgp_perfix_matched_records = bgp_rib.select_prefix(prefix)
    derived = []
    for rec in bgp_perfix_matched_records:
        if rec.Status and rec.Status[0] == "BEST":
            derived.append(BgpRouteNode.from_rec(rec))
        #if rec["Status"] and rec["Status"][0] == "BEST":
        #    derived.append(BgpRouteNode.from_row(rec))

    if len(derived) == 0:
        logging.getLogger(__name__).warning(f"ERROR: missing pred bgp route for {node}")
    return derived, derived, derived

def state_transition_sent_ra_from_bgp(node: DNode):
    if "init" in node.status:
        node.status.remove("init")
    if "sent_by_known_device" in node.status:
        node.status.remove("sent_by_known_device")

lazy_sent_ra_from_bgp = LazyTemplate(
    name="sent_ra_from_bgp",
    matcher=matcher_sent_ra_from_bgp,
    worker=worker_sent_ra_from_bgp,
    state_transition=state_transition_sent_ra_from_bgp,
    edge_type="conj"
)

def matcher_route_from_trp_traced_config(node: DNode) -> bool:
    if isinstance(node, BgpAnnouncementNode) and "trp_finished" in node.status and node.trace:
        return True
    if isinstance(node, BgpRouteNode) and "trp_finished" in node.status and node.trace:
        return True
    return False

def worker_route_from_trp_traced_config(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    policy_specific_traces = defaultdict(list)
    for traceTree in node.trace:
        stack = [traceTree]
        while stack:
            cur = stack.pop()
            trace = cur.traceElement

            # get route-map name from trace element
            rm_name = get_policy_name(trace)
            policy_specific_traces[rm_name].append(trace)
            
            for child in cur.children:
                if child:
                    stack.appendleft(child)

    derived = []
    for rm_name, trace_elements in policy_specific_traces.items():
        if isinstance(node, BgpAnnouncementNode):
            location = node.sender
        elif isinstance(node, BgpRouteNode):
            location = node.host
        device = network.devices[location]
        rm = device.get_routemap(rm_name)
        if rm == None:
            logging.getLogger(__name__).warning(f"WARNING: missing route-map {rm_name} at {location}")
            continue
        #match_node = MatchRoutemapNode(node.pred_route, rm, trace_elements)
        # policy
        rm_node = RoutemapNode(rm.host, "routemap", rm.name, rm.lines)
        derived.append(rm_node)

        # policy term
        for trace in trace_elements:
            cl_type, cl_name = convert_trace_element(trace)
            cl = rm.get_clause(cl_name)
            if cl == None:
                logging.getLogger(__name__).warning(f"WARNING: missing clause for trace: {trace}, cl_type={cl_type}, cl_name={cl_name}")
                continue
            cl_node = RoutemapClauseNode(cl.host, "routemap_clause", cl.name, cl.seq, cl.lines)
            derived.append(cl_node)

            # referenced config elements
            for line_number in cl.lines.lines:
                if line_number in device.referenced_lines:
                    config = device.referenced_lines[line_number]
                    config_node = ReferencedConfigNode.from_config(config)
                    derived.append(config_node)
    return derived, derived, derived

def state_transition_route_from_trp_traced_config(node: DNode):
    if "trp_finished" in node.status:
        node.status.remove("trp_finished")

lazy_route_from_trp_traced_config = LazyTemplate(
    name="route_from_trp_traced_config",
    matcher=matcher_route_from_trp_traced_config,
    worker=worker_route_from_trp_traced_config,
    state_transition=state_transition_route_from_trp_traced_config,
    edge_type="conj"
)

def matcher_sent_ra_from_bgp_session(node: DNode) -> bool:
    if isinstance(node, BgpAnnouncementNode) and "init" in node.status and "sent_by_known_device" in node.status:
        return True
    return False

def worker_sent_ra_from_bgp_session(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    node: BgpAnnouncementNode = node
    sender_name = node.sender
    bgp_edge = node.bgp_edge

    sender = network.devices[sender_name]
    sender_session = sender.find_bgp_session_with_as_ip(bgp_edge.receiver_as, bgp_edge.receiver_ip)

    if sender_session is None:
        logging.getLogger(__name__).warning(f"ERROR: missing bgp session between {bgp_edge.sender}.{bgp_edge.sender_vrf} and {bgp_edge.receiver}.{bgp_edge.receiver_vrf}")
        return [], [], []

    session_node = EstablishedBgpSessionNode.from_session(sender_session, bgp_edge.receiver_vrf)
    if session_node is None:
        logging.getLogger(__name__).warning(f"ERROR: unexpected bgp session status between {bgp_edge.sender}.{bgp_edge.sender_vrf} and {bgp_edge.receiver}.{bgp_edge.receiver_vrf}: {sender_session.established_status}")
        return [], [], []

    derived = [session_node]
    return derived, derived, derived

def state_transition_sent_ra_from_bgp_session(node: DNode):
    if "init" in node.status:
        node.status.remove("init")
    if "sent_by_known_device" in node.status:
        node.status.remove("sent_by_known_device")

lazy_sent_ra_from_bgp_session = LazyTemplate(
    name="sent_ra_from_bgp_session",
    matcher=matcher_sent_ra_from_bgp_session,
    worker=worker_sent_ra_from_bgp_session,
    state_transition=state_transition_sent_ra_from_bgp_session,
    edge_type="conj"
)

def matcher_bgp_from_border_bgp_session(node: DNode) -> bool:
    if isinstance(node, BgpRouteNode) and "received_from_external" in node.status:
        return True
    return False

def worker_bgp_from_border_bgp_session(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    node: BgpRouteNode = node
    session = node.from_session
    session_node = EstablishedBgpSessionNode.from_session(session, "default")
    if not session_node:
        logging.getLogger(__name__).warning(f"ERROR: unexpected bgp session status between {session.host}.{session.vrf} and {session.peer}: {session.established_status}")
        return [], [], []
    derived = [session_node]
    return derived, derived, derived

def state_transition_bgp_from_border_bgp_session(node: DNode):
    if "received_from_external" in node.status:
        node.status.remove("received_from_external")

lazy_bgp_from_border_bgp_session = LazyTemplate(
    name="bgp_from_border_bgp_session",
    matcher=matcher_bgp_from_border_bgp_session,
    worker=worker_bgp_from_border_bgp_session,
    state_transition=state_transition_bgp_from_border_bgp_session,
    edge_type="conj"
)

def matcher_bgp_session_from_peer_config(node: DNode) -> bool:
    if isinstance(node, EstablishedBgpSessionNode) and "init" in node.status:
        return True
    return False

def worker_bgp_session_from_peer_config(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    matched_configs = []
    node: EstablishedBgpSessionNode = node

    if node.is_border:
        vrf = network.get_vrf(node.host, node.vrf)
        remote_as = node.remote_as
        remote_ip = node.remote_ip
        session = vrf.find_bgp_session_with_as_ip(remote_as, remote_ip)
        config = vrf.find_bgp_peer_config_for_session(session)

        if not config:
            logging.getLogger(__name__).warning(f"ERROR: missing bgp peer config at {node.host} for {node.peer}")
            return [], [], []
        matched_configs.append(config)
    else:
        sender_vrf = network.get_vrf(node.host, node.vrf)
        session_as = node.remote_as
        session_ip = node.remote_ip
        sender_session = sender_vrf.find_bgp_session_with_as_ip(session_as, session_ip)
        sender_config = sender_vrf.find_bgp_peer_config_for_session(sender_session)

        if not sender_config:
            logging.getLogger(__name__).warning(f"ERROR: missing bgp peer config at {node.host}.{node.vrf} for {node.peer}.{node.peer_vrf}")
            return [], [], []
        matched_configs.append(sender_config)
    
        receiver_vrf = network.get_vrf(node.peer, node.peer_vrf)
        session_as = node.local_as
        session_ip = node.local_ip
        receiver_session = receiver_vrf.find_bgp_session_with_as_ip(session_as, session_ip)
        receiver_config = receiver_vrf.find_bgp_peer_config_for_session(receiver_session)

        if not receiver_config:
            logging.getLogger(__name__).warning(f"ERROR: missing bgp peer config at {node.peer}.{node.peer_vrf} for {node.host}.{node.vrf}")
            return [], [], []
        matched_configs.append(receiver_config)

    derived = []
    for config in matched_configs:
        # bgp peer config
        if isinstance(config, BgpPeerConfigP2p):
            config_node = BgpPeerConfigP2pNode.from_config(config)
        else:
            config_node = BgpPeerConfigPassiveNode.from_config(config)
        derived.append(config_node)

        # bgp group config
        vrf = network.get_vrf(config.host, config.vrf)
        group_config = vrf.find_bgp_group_for_peer(config)
        if group_config is not None:
            derived.append(BgpGroupConfigNode.from_config(group_config))

    return derived, derived, derived

lazy_bgp_session_from_peer_config = LazyTemplate(
    name="bgp_session_from_peer_config",
    matcher=matcher_bgp_session_from_peer_config,
    worker=worker_bgp_session_from_peer_config,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

def matcher_interface_from_main(node: DNode) -> bool:
    if isinstance(node, MainRouteNode) and "init" in node.status:
        return True
    return False

def worker_interface_from_main(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    node: MainRouteNode = node
    nexthop = node.nexthop
    vrf = network.get_vrf(node.host, node.vrf)

    derived = []
    if nexthop in vrf.interfaces:
        it = vrf.interfaces[nexthop]
        it_node = InterfaceConfigNode.from_interface(it)
        derived.append(it_node)

    return derived, derived, derived

lazy_interface_from_main = LazyTemplate(
    name="interface_from_main",
    matcher=matcher_interface_from_main,
    worker=worker_interface_from_main,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

def matcher_bgp_session_from_l3_connectivity(node: DNode) -> bool:
    if isinstance(node, EstablishedBgpSessionNode) and "init" in node.status:
        return True
    return False

def worker_bgp_session_from_l3_connectivity(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    derived = []
    node: EstablishedBgpSessionNode = node

    egress_conn = L3ConnectivityNode(node.host, node.peer, node.vrf, node.peer_vrf, node.local_ip, node.remote_ip, "TCP", "179", "179")
    derived.append(egress_conn)

    if not node.is_border:
        ingress_conn = L3ConnectivityNode(node.peer, node.host, node.peer_vrf, node.vrf, node.remote_ip, node.local_ip, "TCP", "179", "179")
        derived.append(ingress_conn)

    # specify algorithm to resolve l3 connectivity
    # for single hops, use lpm
    # for others, use batfish traceroute
    for conn_node in derived:
        if node.session_type != "EBGP_SINGLEHOP":
            conn_node.status.add("need_traceroute")
    return derived, derived, derived

lazy_bgp_session_from_l3_connectivity = LazyTemplate(
    name="bgp_session_from_l3_connectivity",
    matcher=matcher_bgp_session_from_l3_connectivity,
    worker=worker_bgp_session_from_l3_connectivity,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

""" def matcher_l3_bi_connectivity_from_main(node: DNode) -> bool:
    if isinstance(node, L3BiConnectivityNode) and "init" in node.status:
        return True
    return False

def worker_l3_bi_connectivity_from_main(network: Network, node: DNode):
    derived = []
    # try lpm resolving from connected routes
    # works for most ebgp/ibgp single-hop sessions

    # 1 -> 2
    sender_name = node.host1
    sender = network.devices[sender_name]
    dst_ip = node.ip2
    rules = sender.resolve_rib_rules_for_ip(dst_ip)
    for rule in rules:
        main_fr = sender.get_rib("main").frame
        rule_matched_records = find_route_records_in_frame(main_fr, rule)
        if len(rule_matched_records) == 1:
            rec = rule_matched_records[0]
            derived.append(MainRouteNode.from_rec(rec))
        else:
            logging.getLogger(__name__).warning(f"WARNING: expect unique match for {rule}, actual matches: {[MainRouteNode.from_rec(rec) for rec in rule_matched_records]}")
    if len(rules) == 0:
        # TODO: fall back to use traceroute
        logging.getLogger(__name__).warning(f"WARNING: cannot find route for {node} within connected routes")

    # 2 -> 1
    sender_name = node.host2
    sender = network.devices[sender_name]
    dst_ip = node.ip1
    rules = sender.resolve_rib_rules_for_ip(dst_ip)
    for rule in rules:
        main_fr = sender.get_rib("main").frame
        rule_matched_records = find_route_records_in_frame(main_fr, rule)
        if len(rule_matched_records) == 1:
            rec = rule_matched_records[0]
            derived.append(MainRouteNode.from_rec(rec))
        else:
            logging.getLogger(__name__).warning(f"WARNING: expect unique match for {rule}, actual matches: {[MainRouteNode.from_rec(rec) for rec in rule_matched_records]}")
    if len(rules) == 0:
        # TODO: fall back to use traceroute
        logging.getLogger(__name__).warning(f"WARNING: cannot find route for {node} within connected routes")

    return derived, derived, derived

lazy_l3_bi_connectivity_from_main = LazyTemplate(
    name="l3_bi_connectivity_from_main",
    matcher=matcher_l3_bi_connectivity_from_main,
    worker=worker_l3_bi_connectivity_from_main,
    state_transition=LazyTemplate.default_state_transition,
) """

def matcher_l3_connectivity_from_main(node: DNode) -> bool:
    if isinstance(node, L3ConnectivityNode) and "init" in node.status and "need_traceroute" not in node.status:
        return True
    return False

def worker_l3_connectivity_from_main(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    derived = []
    node: L3ConnectivityNode = node
    # try lpm resolving from connected routes
    # works for most ebgp/ibgp single-hop sessions

    sender_vrf = network.get_vrf(node.host1, node.vrf1)
    dst_ip = node.ip2
    rules = sender_vrf.resolve_rib_rules_for_ip(dst_ip)
    for rule in rules:
        #main_fr = sender_vrf.get_rib("main").frame
        #rule_matched_records = find_route_records_in_frame(main_fr, rule)
        main_rib = sender_vrf.get_rib("main")
        rule_matched_records = main_rib.select_rule(rule)
        if len(rule_matched_records) == 1:
            rec = rule_matched_records[0]
            derived.append(MainRouteNode.from_rec(rec))
        else:
            logging.getLogger(__name__).warning(f"WARNING: expect unique match for {rule}, actual matches: {[MainRouteNode.from_rec(rec) for rec in rule_matched_records]}")
    if len(rules) == 0:
        # TODO: fall back to use traceroute
        logging.getLogger(__name__).warning(f"WARNING: cannot find route for {node} within connected routes")

    return derived, derived, derived

lazy_l3_connectivity_from_main = LazyTemplate(
    name="l3_connectivity_from_main",
    matcher=matcher_l3_connectivity_from_main,
    worker=worker_l3_connectivity_from_main,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

def matcher_l3_connectivity_from_l3_path(node: DNode) -> bool:
    if isinstance(node, L3ConnectivityNode) and "need_traceroute" in node.status and "init" in node.status:
        return True
    return False

def worker_l3_connectivity_from_l3_path(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    derived = []
    node: L3ConnectivityNode = node
    
    start_loc = f"@vrf({node.vrf1})&" + node.host1
    tr = network.bf.q.traceroute(startLocation=start_loc, headers=HeaderConstraints(dstIps=node.ip2, srcIps=node.ip1, dstPorts=node.port2, srcPorts=node.port1, ipProtocols=node.ipprocotol)).answer().frame()
    for path in tr.Traces[0]:
        derived.append(L3PathNode(node.ip2, path))
    return derived, derived, derived

def state_transition_l3_connectivity_from_l3_path(node: DNode):
    if "need_traceroute" in node.status:
        node.status.remove("need_traceroute")
    if "init" in node.status:
        node.status.remove("init")

lazy_l3_connectivity_from_l3_path = LazyTemplate(
    name="l3_connectivity_from_l3_path",
    matcher=matcher_l3_connectivity_from_l3_path,
    worker=worker_l3_connectivity_from_l3_path,
    state_transition=state_transition_l3_connectivity_from_l3_path,
    edge_type="disj"
)

def matcher_l3_path_from_tested(node: DNode) -> bool:
    if isinstance(node, L3PathNode) and "init" in node.status:
        return True
    return False

def worker_l3_path_from_tested(network: Network, node: DNode) -> Tuple[List[DNode], List[DNode], List[DNode]]:
    node: L3PathNode = node

    derived = convert_traceroute_path(node.path, node.dst_ip)
    return derived, derived, derived

lazy_l3_path_from_tested = LazyTemplate(
    name="l3_path_from_tested",
    matcher=matcher_l3_path_from_tested,
    worker=worker_l3_path_from_tested,
    state_transition=LazyTemplate.default_state_transition,
    edge_type="conj"
)

all_lazy_templates = [
    lazy_tested_from_main,
    lazy_main_from_bgp,
    lazy_main_from_connected,
    lazy_bgp_from_received_ra,
    lazy_bgp_from_connected,
    lazy_bgp_aggregated_from_subnets,
    lazy_sent_ra_from_bgp,
    lazy_route_from_trp_traced_config,
    lazy_sent_ra_from_bgp_session,
    lazy_bgp_from_border_bgp_session,
    lazy_bgp_session_from_peer_config,
    lazy_request_export_trp_for_ra,
    lazy_request_import_trp_for_bgp,
    lazy_interface_from_main,
    lazy_bgp_session_from_l3_connectivity,
    lazy_l3_connectivity_from_main,
    lazy_l3_connectivity_from_l3_path,
    lazy_l3_path_from_tested,
]

lazy_templates_no_policy = [
    lazy_tested_from_main,
    lazy_main_from_bgp,
    lazy_main_from_connected,
    lazy_bgp_from_received_ra,
    lazy_bgp_from_connected,
    lazy_sent_ra_from_bgp,
    lazy_sent_ra_from_bgp_session,
    lazy_bgp_from_border_bgp_session,
    lazy_bgp_session_from_peer_config,
    lazy_interface_from_main,
    lazy_bgp_session_from_l3_connectivity,
    lazy_l3_connectivity_from_l3_path,
    lazy_l3_path_from_tested,
]


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Optional
from pybatfish.client.session import Session
from pybatfish.client.options import Options
from pybatfish.client.consts import CoordConsts

from .algorithm.converttrace import *
from .coverage import Coverage

class NetCovSession(Session):
    def __init__(
        self,
        host: str = Options.coordinator_host,
        port_v1: int = Options.coordinator_work_port,
        port_v2: int = Options.coordinator_work_v2_port,
        ssl: bool = Options.use_ssl,
        verify_ssl_certs: bool = Options.verify_ssl_certs,
        api_key: str = CoordConsts.DEFAULT_API_KEY,
        load_questions: bool = True,
    ):
        super().__init__(host, port_v1, port_v2, ssl, verify_ssl_certs, api_key, load_questions)
        self.cov: Optional[Coverage] = None
    
    # override
    def get_answer(self, question, snapshot, reference_snapshot=None):
        answer = super().get_answer(question, snapshot, reference_snapshot)

        # hook: collect test trace
        if self.cov is not None:
            self.cov.collect_trace(answer)
        return answer

    # override
    def init_snapshot(self, upload, name=None, overwrite=False, extra_args=None, enable_cov=True, static_analysis=False, prebuilt_model=None):
        ss_name = super().init_snapshot(upload, name, overwrite, extra_args)

        if enable_cov:
            self.cov = Coverage(self, upload, static_analysis, prebuilt_model)
        return ss_name

    
        



#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import warnings
from typing import Iterable
import pickle
import os
from dd.autoref import BDD, Function

from ..datamodel import *
from .utils import *


# Generic coverage algorithm given a set of tested nodes
# Assume dependency graph built before call
# Walk the dependency graph
def control_plane_coverage(network: Network, tested_nodes: Iterable[DNode]) -> SourceLines:
    covered_nodes = set()

    def dfs(node: DNode):
        if node in covered_nodes:
            return
        covered_nodes.add(node)
        for (pred, _) in node.pred:
            dfs(pred)

    # assume dependency graph has been built
    # walk dependency graph
    for node in tested_nodes:
        dfs(node)
    
    # stats
    covered_lines = line_level_stats(covered_nodes)
    log_metrics(covered_lines, network, "Configuration coverage")
    return covered_lines

def weak_coverage(network: Network, tested_nodes: Iterable[DNode], metrics: List[str], enable_stats: bool = False) -> SourceLines:
    covered_nodes = set()
    visited = set()
    config_nodes = set()
    precluded_strong_nodes = set()
    
    # discover all config nodes
    def dfs_find_all(node: DNode):
        if node in covered_nodes:
            return

        covered_nodes.add(node)
        if isinstance(node, ConfigNode):
            config_nodes.add(node)

        for (pred, _) in node.pred:
            dfs_find_all(pred)

    for node in tested_nodes:
        dfs_find_all(node)

    # remove all disjunctive edges, config nodes that remain connected are
    # obviously strong.
    def dfs_conj_only(node: DNode):
        if node in visited:
            return

        visited.add(node)
        if isinstance(node, ConfigNode):
            precluded_strong_nodes.add(node)

        for (pred, is_disj) in node.pred:
            if not is_disj:
                dfs_conj_only(pred)
    
    for node in tested_nodes:
        dfs_conj_only(node)

    # assign a Boolean variable for each potentially weak config node
    weak_candicates = config_nodes - precluded_strong_nodes

    bdd = BDD()
    bdd_vars = {}
    predicates = {}
    for i, node in enumerate(weak_candicates):
        var_name = f'c{i}'
        bdd.declare(var_name)
        bdd_vars[node] = var_name
        predicates[node] = bdd.var(var_name)

    # build BDD predicates with post-order traversal
    def dfs_build_predicate(node: DNode) -> Function:
        if node in predicates:
            return predicates[node]

        conj_parents = []
        disj_parents = []
        for (parent, edge_type) in node.pred:
            # avoid precluded config nodes
            if parent in precluded_strong_nodes:
                continue
            parent_pred = dfs_build_predicate(parent)
            if edge_type == False:
                conj_parents.append(parent_pred)
            else:
                disj_parents.append(parent_pred)
        
        if len(disj_parents) == 0:
            res = bdd.true
        else:
            res = bdd.false
            for parent in disj_parents:
                res = res | parent
        for parent in conj_parents:
            res = res & parent
        predicates[node] = res
        return res
        
    for node in tested_nodes:
        dfs_build_predicate(node)

    def dfs_test_weak_coverage(node: DNode, root: DNode, visited: Set[DNode]):
        if node in visited:
            return

        visited.add(node)
        if isinstance(node, ConfigNode):
            # test whether node is weak/strong coverage of root
            if node in weak_candicates and is_strong_map[node]:
                f = predicates[root]
                c = bdd_vars[node]
                #is_strong = f.restrict({c: 0}).is_zero()
                is_strong = bdd.let({c: False}, f) == bdd.false
                is_strong_map[node] &= is_strong

        for (pred, _) in node.pred:
            dfs_test_weak_coverage(pred, root, visited)

    is_strong_map = defaultdict(lambda: True)
    for node in tested_nodes:
        visited = set()
        dfs_test_weak_coverage(node, node, visited)

    #res = {node for node, is_strong in is_strong_map.items() if not is_strong}
    weak_nodes = set()
    for node, is_strong in is_strong_map.items():
        if isinstance(node, ConfigNode) and not is_strong:
            weak_nodes.add(node)

    # stat
    weak_lines = line_level_stats(weak_nodes)

    if "strong" in metrics:
        all_covered_lines = line_level_stats(covered_nodes)
        log_metrics(all_covered_lines.diff(weak_lines), network, "Strong coverage")
    if "weak" in metrics:
        log_metrics(weak_lines, network, "Weak coverage")
    
    if enable_stats:
        logger = logging.getLogger(__name__)
        logger.critical("BDD statistics:")
        
        warnings.filterwarnings("ignore")
        stats = bdd.statistics()
        keys = set(["n_vars", "n_nodes", "peak_nodes", "mem"])
        for key in keys:
            logger.critical(f"{key}: {stats[key]}")
    # force python garbage collection of BDD nodes
    bdd_vars.clear()
    predicates.clear()
    return weak_lines


def print_covered_config_elements(covered_nodes: List[DNode]): 
    covered_bgp_peers = set()
    covered_interfaces = set()
    covered_prefix_lists = set()
    covered_policies = set()
    for node in covered_nodes:
        if isinstance(node, ReferencedConfigNode) and node.config_type == "prefix-list":
            covered_prefix_lists.add(node)
        if isinstance(node, RoutemapNode):
            covered_policies.add(node)
        elif isinstance(node, BgpPeerConfigNode):
            covered_bgp_peers.add(node)
        elif isinstance(node, InterfaceConfigNode):
            covered_interfaces.add(node)
    
    print("Covered Routemaps:")
    for node in covered_policies:
        print(f"  {node} {node.lines}")
    print("Covered Prefix Lists:")
    for node in covered_prefix_lists:
        print(f"  {node} {node.lines}")

def log_metrics(covered_sources: SourceLines, network: Network, metric_name: str="Configuratio coverage", denominator: str="reachable") -> None:
    if denominator == "reachable":
        denom = network.reachable_source
    elif denominator == "supported":
        denom = network.supported_source
    else:
        raise NotImplemented
    # sanity unreachable
    covered_sources = covered_sources.intersect(denom)
    
    #cnt_all = network.source.count()
    cnt_covered = covered_sources.count()
    #cnt_supported = network.supported_source.count()
    cnt_denom = denom.count()
    logger = logging.getLogger(__name__)
    logger.critical(f"{metric_name}:")
    logger.critical(f"    Covered lines:                         {fraction_repr(cnt_covered, cnt_denom)}")
    #logger.critical(f"    Not marked as dead:                    {fraction_repr(cnt_reachable, cnt_reachable)}")
    #logger.critical(f"    Modeled by NetCov:                     {fraction_repr(cnt_supported, cnt_reachable)}")
    #logger.critical(f"    Modeled by Batfish:                    {fraction_repr(cnt_all, cnt_reachable)}")
    
    #logger.critical(f"Element specific coverage:")
    #logger.critical(f"    Policy term:                           {fraction_repr(cnt_covered_clauses, network.cnt_routemap_clauses)}")
    #logger.critical(f"    Bgp peer config:                       {fraction_repr(cnt_covered_bgp_peers, network.cnt_bgp_peer_configs)}")
    #logger.critical(f"    Interface config:                      {fraction_repr(cnt_covered_interfaces, network.cnt_interface)}")

    logger.critical(f"Breakdown:")
    for config_type, type_sources in network.typed_source.items():
        if config_type not in SUPPORTED_CONFIG_TYPES:
            continue
        type_sources = type_sources.intersect(denom)
        cnt_type_sources = type_sources.count()
        typed_covered_sources = covered_sources.intersect(type_sources)
        cnt_covered = typed_covered_sources.count()
        logger.critical(f"    {(config_type + ':').ljust(38)} {fraction_repr(cnt_covered, cnt_type_sources)}")

    # logger.warning(f"Unsupported:")
    # for config_type, sources in network.typed_source.items():
    #     if config_type in SUPPORTED_CONFIG_TYPES:
    #         continue
    #     cnt_all = sources.count()
    #     logger.warning(f"    {(config_type + ':').ljust(38)} {cnt_all}")

def line_level_stats(covered_nodes: Iterable[DNode]) -> SourceLines:
    covered_sources = SourceLines()
    for node in covered_nodes:
        if isinstance(node, ConfigNode):
            if node.lines:
                covered_sources.add_source_lines(node.lines)
    return covered_sources

def bgp_group_breakdown(network: Network, covered_sources: SourceLines, plot_format=False) -> None:
    all_lines: DefaultDict[Tuple(str, str), SourceLines] = defaultdict(SourceLines)
    labels = []
    row1 = []
    row2 = []

    for device_name, vrf_name, device, vrf in network.iter_vrfs():
        for group_name, group in vrf.bgp_group_configs.items():
            all = all_lines[(vrf_name, group_name)]
            for peer in group.peer_configs:
                all.add_source_lines(peer.lines)
    
    print(f"BGP peer coverage breakdown to groups:")
    for (vrf_name, group_name), all in all_lines.items():
        cnt_all = all.count()
        cnt_covered = all.intersect(covered_sources).count()
        if cnt_all != 0:
            print(f"  {vrf_name.ljust(8)}{group_name.ljust(15)}: {cnt_covered}/{cnt_all} ({'{:.2%}'.format(cnt_covered/cnt_all) if cnt_all != 0 else '0.00%'})")
            if plot_format and vrf_name == "default":
                labels.append(group_name)
                row1.append(cnt_covered)
                row2.append(cnt_all)

    if plot_format:
        a = [(labels[i], row1[i], row2[i]) for i in range(len(labels))]
        a.sort(key = lambda x: x[2], reverse=True)
        labels = [x[0] for x in a]
        row1 = [x[1] for x in a]
        row2 = [x[2] for x in a]
        print(labels)
        print(row1)
        print(row2)


def data_plane_coverage(network: Network, tested_nodes: Iterable[DNode]) -> None:
    covered_rules = set()
    covered_interfaces = set()

    def dfs(node: DNode, depth = 1):
        if depth > 3:
            return

        # metric stat
        if isinstance(node, MainRouteNode):
            covered_rules.add(node)
        if isinstance(node, InterfaceConfigNode):
            covered_interfaces.add(node)
         
        for (pred, _) in node.pred:
            dfs(pred, depth + 1)

    # assume dependency graph has been built

    # walk dependency graph
    for node in tested_nodes:
        dfs(node)

    print(f"")
    print(f"Dataplane coverage:")
    print(f"    Rule coverage:      {len(covered_rules)}/{network.cnt_rib_entry} ({'{:.2%}'.format(len(covered_rules)/network.cnt_rib_entry)})")
    print(f"    Interface Coverage: {len(covered_interfaces)}/{network.cnt_interface} ({'{:.2%}'.format(len(covered_interfaces)/network.cnt_interface)})")


def pickle_tested_nodes(tested_nodes: Iterable[DNode], filename) -> None:
    with open(filename, 'wb') as outfile:
        pickle.dump(tested_nodes, outfile)

def unpickle_tested_nodes(filename: str) -> List[DNode]:
    with open(filename, 'rb') as infile:
        return pickle.load(infile)

def pickle_covered_sources(covered: SourceLines, filename: str) -> None:
    with open(filename, 'wb') as outfile:
        pickle.dump(covered, outfile)

def unpickle_covered_sources(filename: str) -> SourceLines:
    with open(filename, 'rb') as infile:
        return pickle.load(infile)

def dump_lcov(covered: SourceLines, universe: SourceLines, lcovfile: str, uncovered:bool = False) -> None:
    os.makedirs(os.path.dirname(lcovfile), exist_ok=True)
    with open(lcovfile, 'w') as outfile:
        if uncovered: # highlight both covered and uncovered
            for filename, lines in universe.files2lines.items():
                covered_lines = covered.files2lines[filename]
                lcov_lines = ["TN:\n", f"SF:./{filename}\n", f"LF:{len(lines)}\n", f"LH:{len(covered_lines)}\n"]
                for line in lines:
                    if isinstance(line, int):
                        if line in covered_lines:
                            lcov_lines.append(f"DA:{line},1\n")
                        else:
                            lcov_lines.append(f"DA:{line},0\n")
                lcov_lines.append("end_of_record\n")
                outfile.writelines(lcov_lines)
        else: # highlight covered lines only
            for filename, lines in covered.files2lines.items():
                lcov_lines = ["TN:\n", f"SF:./{filename}\n", f"LF:{len(universe.files2lines[filename])}\n", f"LH:{len(lines)}\n"]
                for line in lines:
                    if isinstance(line, int):
                        lcov_lines.append(f"DA:{line},1\n")
                lcov_lines.append("end_of_record\n")
                outfile.writelines(lcov_lines)


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import annotations # type hint in the enclosing class
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Set, List
from pybatfish.datamodel.primitives import FileLines

class SourceLines:
    def __init__(self) -> None:
        self.files2lines: DefaultDict[str, Set[int]] = defaultdict(set)
        
    def add_source_lines(self, fileLines: FileLines):
        for line in fileLines.lines:
            self.files2lines[fileLines.filename].add(line)

    def add_line(self, filename: str, line: int) -> None:
        self.files2lines[filename].add(line)

    def intersect(self, other: SourceLines) -> SourceLines:
        res = SourceLines()
        for filename in self.files2lines.keys():
            if filename in other.files2lines:
                for line in self.files2lines[filename]:
                    if line in other.files2lines[filename]:
                        res.add_line(filename, line)
        return res

    def update(self, other: SourceLines) -> None:
        for filename, lines in other.files2lines.items():
            for line in lines:
                self.files2lines[filename].add(line)

    def diff(self, other: SourceLines) -> SourceLines:
        """a.diff(b) = elements in a but not in b
        """
        res = SourceLines()
        for filename in self.files2lines.keys():
            if filename in other.files2lines:
                for line in self.files2lines[filename]:
                    if line not in other.files2lines[filename]:
                        res.add_line(filename, line)
            else:
                for line in self.files2lines[filename]:
                    res.add_line(filename, line)
        return res
    
    def contains(self, fileLines: FileLines) -> bool:
        all_contained = True
        for line in fileLines.lines:
            if line not in self.files2lines[fileLines.filename]:
                all_contained = False
                break
        return all_contained

    def select_files(self, filenames: List[str]) -> SourceLines:
        res = SourceLines()
        for filename in filenames:
            if filename in self.files2lines:
                res.files2lines[filename].update(self.files2lines[filename])
        return res
        
    def from_dict(d: Dict) -> SourceLines:
        res = SourceLines()
        for filename, lines in d.items():
            res.files2lines[filename].update(lines)
        return res

    def count(self) -> int:
        cnt = 0
        for lines in self.files2lines.values():
            cnt += len(lines)
        return cnt

    def to_filelines(self, filename) -> FileLines:
        return FileLines(filename=filename, lines=list(self.files2lines[filename]))

    def print(self) -> str:
        reprs = []
        for filename, lines in self.files2lines.items():
            if len(lines) > 0:
                reprs.append(f"{filename}({', '.join([str(line) for line in lines])})")
        return '\n'.join(reprs)



#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from .utils import get_equivalent
from .dnode import DNode


class IFG:
    def __init__(self) -> None:
        self.nodes = set()

    def add_node(self, node: DNode) -> None:
        self.nodes.add(node)

    def get_node(self, node: DNode) -> DNode:
        return get_equivalent(self.nodes, node)



#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
from ..datamodel import *
from .utils import *



def convert_traceroute_traces(trs: Iterable[pd.DataFrame], print_stats=False) -> List[DNode]:
    tested_nodes = set()
    action_stats = defaultdict(int)
    tr_cnt = 0
    path_cnt = 0
    for tr in trs:
        tr_cnt += 1
        dst_ip = tr.Flow[0].dstIp
        for path in tr.Traces[0]:
            path_cnt += 1
            action_stats[path[-1][-1]] += 1

            tested_nodes.update(convert_traceroute_path(path, dst_ip))

    if print_stats:
        print("Traceroute stats:")
        print(f"  Count of traceroutes: {tr_cnt}")
        print(f"  Count of paths:       {path_cnt}")
        for action, cnt in action_stats.items():
            print(f"      {action} : {cnt}")
            
    return list(tested_nodes)

def convert_trp_traces(network: Network, trps: List[pd.DataFrame], filter_by_action_enabled: bool=False, desired_action: str='') -> List[DNode]:
    tested_nodes = set()
    logger = logging.getLogger(__name__)

    for trp in trps:
        for i in range(len(trp.index)):
            trp_record = trp.iloc[i]
            if filter_by_action_enabled:
                if trp_record.Action != desired_action:
                    continue
            input_route = trp_record.Input_Route
            traces = trp_record.Trace
            host_name = trp_record.Node
            policy_specific_traces = defaultdict(list)
            for traceTree in traces:
                stack = [traceTree]
                while stack:
                    cur = stack.pop()
                    trace = cur.traceElement

                    # get route-map name from trace element
                    rm_name = get_policy_name(trace)
                    policy_specific_traces[rm_name].append(trace)
                    
                    for child in cur.children:
                        if child:
                            stack.appendleft(child)

            device = network.devices[host_name]
            for rm_name, trace_elements in policy_specific_traces.items():
                rm = device.get_routemap(rm_name)
                if rm == None:
                    logger.warning(f"Missing route-map {rm_name} at {host_name}")
                    continue
                rm_node = RoutemapNode(rm.host, "routemap", rm.name, rm.lines)
                tested_nodes.add(rm_node)

                # policy term
                for trace in trace_elements:
                    cl_type, cl_name = convert_trace_element(trace)
                    cl = rm.get_clause(cl_name)
                    if cl == None:
                        logger.warning(f"Missing clause for trace: {trace}, cl_type={cl_type}, cl_name={cl_name}")
                        continue
                    cl_node = RoutemapClauseNode(cl.host, "routemap_clause", cl.name, cl.seq, cl.lines)
                    tested_nodes.add(cl_node)

                    # referenced config elements
                    for line_number in cl.lines.lines:
                        if line_number in device.referenced_lines:
                            config = device.referenced_lines[line_number]
                            config_node = ReferencedConfigNode.from_config(config)
                            tested_nodes.add(config_node)
    return list(tested_nodes)

def convert_main_rib_routes(routes: pd.DataFrame) -> List[DNode]:
    tested_nodes = set()
    for rec in routes.to_records():
        tested_nodes.add(MainRouteNode.from_rec(rec))
    return list(tested_nodes)

def convert_bgp_routes(routes: pd.DataFrame) -> List[DNode]:
    tested_nodes = set()
    for rec in routes.to_records():
        tested_nodes.add(BgpRouteNode.from_rec(rec))
    return list(tested_nodes)

def convert_raw_config(config: Dict) -> List[DNode]:
    file_lines = FileLines(filename=config['filename'], lines=config['lines'])
    config_type = config['type'] if 'type' in config else 'user-supplied'
    node = UserSuppliedConfigNode(config['host'], file_lines, config_type)
    return node


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import re
import random

from .utils import *
from .converttrace import *
from ..datamodel.network import *
from ..datamodel.dnode import *

def test_case_single_fib_rule(device_name, prefix, next_hop_ip, next_hop_interface) -> Iterable[DNode]:
    return [DataplaneTestNode("forwarded_to_ip", device_name, prefix, next_hop_ip, next_hop_interface)]

def test_case_single_device(network: Network, device_name: str):
    device = network.devices[device_name]
    nodes = []
    main_records = device.get_rib("main").frame.to_records()

    for rec in main_records:
        nodes.append(MainRouteNode.from_rec(rec))
    return nodes

def test_case_single_device_bgp(network: Network, device_name: str):
    vrf = network.get_vrf(device_name, "default")
    nodes = []
    main_records = vrf.get_rib("main").frame.to_records()

    for rec in main_records:
        if rec.Protocol in ["bgp", "ibgp"]:
            nodes.append(MainRouteNode.from_rec(rec))
    return nodes

def test_case_single_prefix(network: Network, prefix: str):
    nodes =[]
    for device_name, device in network.devices.items():
        main_records = device.get_rib("main").frame.to_records()

        for rec in main_records:
            if rec.Network == prefix:
                nodes.append(MainRouteNode.from_rec(rec))

    return nodes


def test_case_full_dataplane(bf) -> Iterable[DNode]:
    nodes =[]
    routes: pd.DataFrame = bf.q.routes().answer().frame()

    for rec in routes.to_records():
        node = MainRouteNode.from_rec(rec)
        if node != None:
            nodes.append(node)

    return nodes


def test_case_internet2_lb_reach(network: Network) -> Iterable[DNode]:
    # collect loopback address
    devices = network.devices.keys()
    lb_map = {}
    for device_name in devices:
        ifp = network.bf.q.interfaceProperties(nodes=device_name, interfaces="lo0.0").answer().frame()
        lb_map[device_name] = ifp.iloc[0]["Primary_Address"]

    # N-by-N traceroute
    trs = []
    for device_name in devices:
        dst_ip = lb_map[device_name]
        for other_device in devices:
            if other_device == device_name:
                continue
            start_loc = "@vrf(default)&" + other_device
            src_ip = lb_map[other_device]
            tr = network.bf.q.traceroute(startLocation=start_loc, headers=HeaderConstraints(dstIps=dst_ip, srcIps=src_ip)).answer().frame()
            trs.append(tr)

    return convert_traceroute_traces(trs)

def test_case_internet2_lb_to_all_connected(network: Network) -> Iterable[DNode]:
    # collect loopback address
    devices = network.devices.keys()
    lb_map = {}
    for device_name in devices:
        ifp = network.bf.q.interfaceProperties(nodes=device_name, interfaces="lo0.0").answer().frame()
        lb_map[device_name] = ifp.iloc[0]["Primary_Address"]

    # collect all connected routes (in default VRFs)
    cr_map = defaultdict(list)
    for device_name in devices:
        ifp = network.bf.q.interfaceProperties(nodes=device_name).answer().frame()
        crs = cr_map[device_name]
        for rec in ifp.to_records():
            if rec.VRF != "default":
                continue
            prefix = rec.Primary_Address
            if prefix != None:
                crs.append(prefix)

    # N-by-N-by-X traceroute
    trs = []
    test_results = []
    for device_name in devices:
        for prefix in cr_map[device_name]:
            dst_ip = prefix
            for other_device in devices:
                if other_device == device_name:
                    continue
                start_loc = "@vrf(default)&" + other_device
                src_ip = lb_map[other_device]
                tr = network.bf.q.traceroute(startLocation=start_loc, headers=HeaderConstraints(dstIps=dst_ip, srcIps=src_ip)).answer().frame()
                
                passed = True
                for path in tr.Traces[0]:
                    passed &= path[-1][-1].action in ["DELIVERED_TO_SUBNET", 'ACCEPTED', "EXITS_NETWORK"]
                
                test_results.append((start_loc, dst_ip, passed))
                if passed:
                    trs.append(tr)

    return convert_traceroute_traces(trs)

    
def test_case_internet2_route_pref(network, external_ras, rel_file) -> Iterable[DNode]:
    # load as relation from file
    rel = {}
    with open(rel_file) as infile:
        rel.update(json.load(infile))

    # enumerate sender's asn for all external peers
    prefix_to_as_map = defaultdict(set)
    for ra in external_ras:
        asn = ra["asPath"][0][0]
        if str(asn) in rel:
            prefix_to_as_map[ra["network"]].add(asn)
    
    # find prefixes that are sent from different types of peers
    prefix_to_peer_type_map = {}
    for prefix, ass in prefix_to_as_map.items():
        types = set()
        for asn in ass:
            if str(asn) in rel:
                types.add(rel[str(asn)])
        if len(types) > 1:
            prefix_to_peer_type_map[prefix]=types

    #print(prefix_to_peer_type_map)

    test_results = []
    bgp_fr = network.bf.q.bgpRib(vrfs = "default", status="/.*/").answer().frame()
    tested_nodes = []
    for prefix, types in prefix_to_peer_type_map.items():
        matched_fr = bgp_fr[bgp_fr["Network"] == prefix]
        matched_records = matched_fr.to_records()
        
        pref_map = defaultdict(list)
        for rec in matched_records:
            asn = rec.AS_Path.split(' ')[0]
            if asn in rel:
                pref_map[rel[asn]].append((asn, rec))

        if len(pref_map) > 1:
            #print(prefix, pref_map)
            rels = list(pref_map.keys())
            rels.sort()
            case_passed = True
            n = len(rels)
            for i in range(n-1):
                # routes in i-th bucket should all have lower pref than >i th bucket
                max_i = 0
                for asn, rec in pref_map[rels[i]]:
                    if rec.Local_Pref > max_i:
                        max_i = rec.Local_Pref
                
                min_higher_class = 1000
                for j in range(i+1, n):
                    for asn, rec in pref_map[rels[j]]:
                        if rec.Local_Pref < min_higher_class:
                            min_higher_class = rec.Local_Pref
                case_passed &= max_i < min_higher_class

            test_results.append((case_passed, pref_map))
            if case_passed:
                for rec in matched_records:
                    tested_nodes.append(BgpRouteNode.from_rec(rec))

    return tested_nodes

def test_case_internet2_no_martian(network: Network, border_sessions) -> Iterable[DNode]:
    martians = [
        "0.0.0.0/0",
        "10.0.0.0/8",
        "127.0.0.0/8",
        "169.254.0.0/16",
        "172.16.0.0/12",
        "192.0.2.0/24",
        "192.88.99.1/32",
        "192.168.0.0/16",
        "198.18.0.0/15",
        "224.0.0.0/4",
        "240.0.0.0/4",
    ]

    trps = []
    for s in border_sessions:
        input_routes = [BgpRoute(network=prefix, originatorIp="0.0.0.0", originType="egp", protocol='bgp', asPath=[[s['remote_as']]], communities=[]) for prefix in martians]
        device_name = s['node']
        device = network.devices[device_name]
        policy = f"~PEER_IMPORT_POLICY:{s['remote_ip']}/32~"
        trp_batch_result = network.bf.q.testRoutePolicies(nodes=device_name, policies=policy, inputRoutes=input_routes, direction="in").answer().frame()
        trps.append(trp_batch_result)

    return convert_trp_traces(network, trps, True, 'DENY')

def test_case_internet2_bte(network: Network, sample: int = 1) -> Iterable[DNode]:
    IBGP_GROUPS = {
        "INTERNET2",
        "OTHER-INTERNAL"
    }

    tested_nodes = []

    def sample_routes(bgp_rib: IndexedRib, nmax: int) -> List[BgpRoute]:
        if len(bgp_rib.prefixmap) == 0:
            return []
        rib = np.hstack(bgp_rib.prefixmap.values())

        if rib.size <= nmax:
            sampled = rib
        else:
            sampled = rib[np.random.choice(rib.shape[0], nmax, replace=False)]

        for route in sampled:
            route.Communities = ["11537:888", *route.Communities]
        return [convert_bgp_route(rec) for rec in sampled]

        
    trps = []
    for device_name, vrf_name, device, vrf in network.iter_vrfs():
        if not vrf.bgp_enabled:
            continue
        bgp_rib = vrf.get_rib("bgp")
        input_routes = sample_routes(bgp_rib, sample)
        if not input_routes:
            continue
        for peer_config in vrf.bgp_peer_configs:
            if peer_config.peer_group in IBGP_GROUPS:
                continue
            policy = find_composed_peer_policy(device.raw_policies, peer_config.remote_ip, "EXPORT")
            if policy == None:
                print(f"TEST WARNING: cannot find export policy for peer {peer_config}")
                continue
            trp_batch_result = network.bf.q.testRoutePolicies(nodes=device_name, policies=policy, inputRoutes=input_routes, direction="out").answer().frame()
            trps.append(trp_batch_result)

    return convert_trp_traces(network, trps, True, 'DENY')

def test_case_internet2_sanityin(network: Network) -> Iterable[DNode]:
    internels = [
        "64.57.16.0/20",
        "64.57.22.0/24",
        "64.57.23.240/28",
        "198.32.8.0/22",
        "198.32.12.0/22",
        "198.32.154.0/24",
        "198.71.45.0/24",
        "198.71.46.0/24",
    ]
    forbidden_asns = [
        19401, # NLR
        *[random.randint(64512, 65535) for _ in range(10)], # PRIVATE
        174, # COMMERCIAL
        701,
        1239,
        1673,
        1740,
        1800,
        1833,
        2551,
        2548,
        2685,
        2914,
        3549,
        3561,
        3847,
        3951,
        3967,
        4183,
        4200,
        5683,
        6113,
        6172,
        6461,
        7018
    ]
    tested_nodes = []

    input_routes = []

    # internal test cases
    input_routes.extend([BgpRoute(network=prefix, originatorIp="0.0.0.0", originType="egp", protocol='bgp', asPath=[], communities=[]) for prefix in internels])
    # as path test cases
    input_routes.extend([BgpRoute(network="1.1.1.1/24", originatorIp="0.0.0.0", originType="egp", protocol='bgp', asPath=[[asn]], communities=[]) for asn in forbidden_asns])


    trps = []
    for device_name in network.devices.keys():
        device = network.devices[device_name]
        policy = "SANITY-IN"
        trp_batch_result = network.bf.q.testRoutePolicies(nodes=device_name, policies=policy, inputRoutes=input_routes, direction="in").answer().frame()
        trps.append(trp_batch_result)

    return convert_trp_traces(network, trps, True, 'DENY')

def test_case_internet2_peer_specific_policy(network: Network) -> Iterable[DNode]:
    # {(device_name, policy_name) : [prefix]}
    test_case: DefaultDict[Tuple[str, str, str], Set[str]] = defaultdict(set)

    # collect prefix-list contents
    named_fr = network.bf.q.namedStructures(structureTypes="Route_Filter_List").answer().frame()
    def sample_prefix_from_rec(rec: np.record) -> Optional[str]:
        if 'lines' in rec.Structure_Definition:
            lines = rec.Structure_Definition['lines']
            if len(lines) > 0:
                prefix = lines[0]['ipWildcard']
                if lines[0]['lengthRange'] == '32-32':
                    prefix = prefix + '/32'
                return prefix
        return None

    # collect bgp peers and -IN/-OUT policies
    def is_peer_specific_policy(policy_name: str) -> bool:
        return re.search("^(?!.*(SANITY|CONNECTOR|ITN|COMMS)).*(-IN$|-OUT$)", policy_name) is not None
    
    def get_direction(policy_name: str) -> str:
        return 'in' if re.search("-IN$", policy_name) is not None else 'out'

    for device_name, vrf_name, device, vrf in network.iter_vrfs():
        for bgp_peer_config in vrf.bgp_peer_configs:
            for policy_name in bgp_peer_config.import_policies + bgp_peer_config.export_policies:
                if is_peer_specific_policy(policy_name):
                    policy = device.get_routemap(policy_name)
                    for line in policy.raw_lines.lines:
                        if line in device.referenced_lines:
                            ref = device.referenced_lines[line]
                            if ref.config_type == "prefix-list":
                                matched_records = named_fr[(named_fr["Node"] == device_name) & (named_fr["Structure_Name"] == ref.name)].to_records()
                                if len(matched_records) != 1:
                                    print(f"WARNING: expect unique match for {policy_name} at {device_name}, actual matches: {len(matched_records)}")
                                prefix = sample_prefix_from_rec(matched_records[0])
                                if prefix is not None:
                                    test_case[(device_name, policy_name)].add(prefix)
    
    # test trp and collect traces
    trps = []
    for (host, policy), prefixes in test_case.items():
        input_routes = [BgpRoute(
            network=prefix,
            protocol='BGP',
            asPath=[],
            communities=[],
            originatorIp='0.0.0.0',
            originType='egp',
        ) for prefix in prefixes]
        direction = get_direction(policy)
        device = network.devices[host]

        trp_batch_result = network.bf.q.testRoutePolicies(nodes=host, policies=policy, inputRoutes=input_routes, direction=direction).answer().frame()        
        trps.append(trp_batch_result)

    return convert_trp_traces(network, trps)

def test_case_internet2_allow_in(external_ras: List[Dict], network: Network) -> Iterable[DNode]:
    tested_nodes = []

    named_fr = network.bf.q.namedStructures(structureTypes="Route_Filter_List").answer().frame()
    def extract_prefix_from_rec(rec: np.record) -> Set[str]:
        res = set()
        if 'lines' in rec.Structure_Definition:
            lines = rec.Structure_Definition['lines']
            for line in lines:
                prefix = line['ipWildcard']
                if line['lengthRange'] == '32-32':
                    prefix = prefix + '/32'
                res.add(prefix)
        return res

    # collect bgp peers and -IN/-OUT policies
    def is_peer_specific_policy(policy_name: str) -> bool:
        return re.search("^(?!.*(SANITY|CONNECTOR|ITN|COMMS)).*(-IN$|-OUT$)", policy_name) is not None

    allow_ins: Dict[BgpPeerConfigP2p, Set[str]] = {}
    for device_name, vrf_name, device, vrf in network.iter_vrfs():
        for bgp_peer_config in vrf.bgp_peer_configs:
            peer_allow_in = set()
            for policy_name in bgp_peer_config.import_policies:
                if is_peer_specific_policy(policy_name):
                    policy = device.get_routemap(policy_name)
                    for line in policy.raw_lines.lines:
                        if line in device.referenced_lines:
                            ref = device.referenced_lines[line]
                            if ref.config_type == "prefix-list":
                                matched_records = named_fr[(named_fr["Node"] == device_name) & (named_fr["Structure_Name"] == ref.name)].to_records()
                                if len(matched_records) != 1:
                                    print(f"WARNING: expect unique match for {policy_name} at {device_name}, actual matches: {len(matched_records)}")
                                prefixes = extract_prefix_from_rec(matched_records[0])
                                peer_allow_in.update(prefixes)
            if peer_allow_in:
                allow_ins[bgp_peer_config] = peer_allow_in

    test_results = []
    for ra in external_ras:
        prefix = ra["network"]
        remote_ip = ra["srcIp"]
        for peer_config, allow_in in allow_ins.items():
            if peer_config.remote_ip == remote_ip and prefix in allow_in:
                # test logic: route should exist in vrf's bgp rib
                vrf = network.get_vrf(peer_config.host, peer_config.vrf)
                bgp_rib = vrf.get_rib("bgp")
                case_passed =  prefix in bgp_rib.prefixmap
                test_results.append((prefix, peer_config, case_passed))
                # by definition, only passed test cases will count for coverage
                if case_passed:
                    tested_nodes.extend([BgpRouteNode.from_rec(rec) for rec in bgp_rib.select_prefix(prefix)])
    
    return tested_nodes

def test_case_fattree_pingmesh(network: Network) -> Iterable[DNode]:
    leaves = retrieve_leaves("/edge/")

    trs = []
    for leaf in leaves:
        for other_leaf in leaves:
            if other_leaf == leaf:
                continue
            tr_frame = network.bf.q.traceroute(startLocation=leaf, headers=HeaderConstraints(srcIps=f"{leaf}[Loopback0]", dstIps=f"{other_leaf}[Ethernet1]")).answer().frame()
            trs.append(tr_frame.iloc[[0]])

    tested_nodes = convert_traceroute_traces(trs, print_stats=True)
    return tested_nodes

def test_case_fattree_default_route_check(network: Network) -> Iterable[DNode]:
    routes: pd.DataFrame = network.bf.q.routes(network="0.0.0.0/0").answer().frame()

    tested_nodes = []
    for rec in routes.itertuples():
        tested_nodes.append(MainRouteNode.from_rec(rec))
    return tested_nodes

def test_case_fattree_aggregate(network: Network, ext_sessions: List[Dict]) -> Iterable[DNode]:
    tested_nodes = []

    for border_session in ext_sessions:
        device_name = border_session["node"]
        remote_as = border_session["remote_as"]
        remote_ip = border_session["remote_ip"]
        device = network.devices[device_name]
        session = device.find_bgp_session_with_as_ip(remote_as, remote_ip)
        bgp_edge = network.get_bgp_edge(device_name, "default", session.peer, 'vrf')
        ra_node = BgpAnnouncementNode("route_announcement", "10.0.0.0/8", None, session.host, session.peer,\
            "default", "default", None, bgp_edge, None, None)
        ra_node.status.add("sent_by_known_device")
        tested_nodes.append(ra_node)
    return tested_nodes